# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/02_text.models.core.ipynb.

# %% auto 0
__all__ = ['SequentialRNN', 'SentenceEncoder', 'AttentiveSentenceEncoder', 'masked_concat_pool', 'XPoolingLinearClassifier',
           'LabelAttentionClassifier', 'get_xmltext_classifier', 'awd_lstm_xclas_split', 'get_xmltext_classifier2']

# %% ../../../nbs/02_text.models.core.ipynb 1
from fastai.data.all import *
from fastai.text.models.core import *
from fastai.text.models.awdlstm import *
from ...layers import *

# %% ../../../nbs/02_text.models.core.ipynb 4
_model_meta = {AWD_LSTM: {'hid_name':'emb_sz', 'url':URLs.WT103_FWD, 'url_bwd':URLs.WT103_BWD,
                          'config_lm':awd_lstm_lm_config, 'split_lm': awd_lstm_lm_split,
                          'config_clas':awd_lstm_clas_config, 'split_clas': awd_lstm_clas_split},}

# %% ../../../nbs/02_text.models.core.ipynb 8
class SequentialRNN(nn.Sequential):
    "A sequential pytorch module that passes the reset call to its children."
    def reset(self):
        for c in self.children(): getattr(c, 'reset', noop)()

# %% ../../../nbs/02_text.models.core.ipynb 10
def _pad_tensor(t, bs):
    if t.size(0) < bs: return torch.cat([t, t.new_zeros(bs-t.size(0), *t.shape[1:])])
    return t

# %% ../../../nbs/02_text.models.core.ipynb 12
class SentenceEncoder(Module):
    "Create an encoder over `module` that can process a full sentence."
    def __init__(self, bptt, module, pad_idx=1, max_len=None): store_attr('bptt,module,pad_idx,max_len')
    def reset(self): getattr(self.module, 'reset', noop)()

    def forward(self, input):
        bs,sl = input.size()
        self.reset()
        mask = input == self.pad_idx
        outs,masks = [],[]
        for i in range(0, sl, self.bptt):
            #Note: this expects that sequence really begins on a round multiple of bptt
            real_bs = (input[:,i] != self.pad_idx).long().sum()
            o = self.module(input[:real_bs,i: min(i+self.bptt, sl)])
            if self.max_len is None or sl-i <= self.max_len:
                outs.append(o)
                masks.append(mask[:,i: min(i+self.bptt, sl)])
        outs = torch.cat([_pad_tensor(o, bs) for o in outs], dim=1)
        inps = input[:, -outs.shape[1]:] # the ofsetted tokens for the outs
        mask = torch.cat(masks, dim=1)
        return inps,outs,mask

# %% ../../../nbs/02_text.models.core.ipynb 14
from ...layers import inattention

# %% ../../../nbs/02_text.models.core.ipynb 15
class AttentiveSentenceEncoder(Module):
    "Create an encoder over `module` that can process a full sentence."
    def __init__(self, bptt, module, decoder, pad_idx=1, max_len=None, running_decoder=True): 
        store_attr('bptt,module,decoder,pad_idx,max_len,running_decoder')
        self.n_lbs = getattr(self.decoder, 'n_lbs', None)
        
    def reset(self): 
        getattr(self.module, 'reset', noop)()

    def forward(self, input):
        bs,sl = input.size()
        self.reset()
        self.decoder.hl = input.new_zeros((bs, self.n_lbs))
        # print(f"Starting to read a btch of docs start hl.sum() = {self.decoder.hl.sum()}", end='\n')
        mask = input == self.pad_idx
        outs,masks = [],[]
        for i in range(0, sl, self.bptt):
            #Note: this expects that sequence really begins on a round multiple of bptt
            real_bs = (input[:,i] != self.pad_idx).long().sum()
            chunk = slice(i, min(i+self.bptt, sl))
            o = self.module(input[:real_bs, chunk]) # shape (bs, bptt, nh)
            if self.max_len is None or sl-i <= self.max_len:
                outs.append(o)
                masks.append(mask[:, chunk])
                # print(f"\t\t (Within max_len) After reading bptt chunk: hl.sum() = {self.decoder.hl.sum()}", end='\n')
            elif self.running_decoder:
                mask_slice = mask[:real_bs, chunk] 
                inp = input[:real_bs, chunk]
                hl, *_ = self.decoder((inp, o, mask_slice))

                self.decoder.hl = hl.inattention(k=3, sort_dim=-1).sigmoid().detach()
                # print(f"\t (Outside max_len) After reading bptt chunk: hl.sum() = {self.decoder.hl.sum()}", end='\n')
                
        # import pdb; pdb.set_trace()
        outs = torch.cat([_pad_tensor(o, bs) for o in outs], dim=1)
        inps = input[:, -outs.shape[1]:] # the ofsetted tokens for the outs
        mask = torch.cat(masks, dim=1)
        return inps, outs, mask

# %% ../../../nbs/02_text.models.core.ipynb 16
def masked_concat_pool(output, mask, bptt):
    "Pool `MultiBatchEncoder` outputs into one vector [last_hidden, max_pool, avg_pool]"
    lens = output.shape[1] - mask.long().sum(dim=1)
    last_lens = mask[:,-bptt:].long().sum(dim=1)
    avg_pool = output.masked_fill(mask[:, :, None], 0).sum(dim=1)
    avg_pool.div_(lens.type(avg_pool.dtype)[:,None])
    max_pool = output.masked_fill(mask[:,:,None], -float('inf')).max(dim=1)[0]
    x = torch.cat([output[torch.arange(0, output.size(0)),-last_lens-1], max_pool, avg_pool], 1) #Concat pooling.
    return x

# %% ../../../nbs/02_text.models.core.ipynb 17
class XPoolingLinearClassifier(Module):
    def __init__(self, dims, ps, bptt, y_range=None):
        self.layer = LinBnDrop(dims[0], dims[1], p=ps, act=None)
        self.bptt = bptt

    def forward(self, input):
        out, mask = input
        x = masked_concat_pool(out, mask, self.bptt)
        x = self.layer(x)
        return x, out, out

# %% ../../../nbs/02_text.models.core.ipynb 20
from ...layers import _create_bias

# %% ../../../nbs/02_text.models.core.ipynb 21
from ...utils import *

# %% ../../../nbs/02_text.models.core.ipynb 22
class LabelAttentionClassifier(Module):
    initrange=0.1
    def __init__(self, n_hidden, n_lbs, plant=0.5, attn_init=(0,0,1), y_range=None, static_inattn=5, diff_inattn=30, lowshot=False, unfreeze_lm_decoder=False):
        store_attr('n_hidden,n_lbs,y_range')
        self.pay_attn = XMLAttention(self.n_lbs, self.n_hidden, plant=plant, attn_init=attn_init, static_inattn=static_inattn, diff_inattn=diff_inattn, lowshot=lowshot, unfreeze_lm_decoder=unfreeze_lm_decoder)
        # self.boost_attn = ElemWiseLin(self.n_lbs, self.n_hidden)
        self.boost_attn = ElemWiseLin(self.n_lbs, 400)
        self.label_bias = _create_bias((self.n_lbs,), with_zeros=False)
        self.hl = torch.zeros(1)
    
    def forward(self, sentc):
        if isinstance(sentc, tuple): inp, sentc, mask = sentc # sentc is the stuff coming outta SentenceEncoder i.e., shape (bs, max_len, nh) in other words the concatenated output of the AWD_LSTM
        test_eqs(inp.shape, sentc.shape[:-1], mask.shape)
        sentc = sentc.masked_fill(mask[:, :, None], 0)
        # import pdb; pdb.set_trace()
        attn, wgts, lbs_cf, lm_loss = self.pay_attn(inp, sentc, mask) #shape (bs, n_lbs, n_hidden)
        attn = self.boost_attn(attn) # shape (bs, n_lbs, n_hidden)
        bs = self.hl.size(0)
        self.hl = self.hl.to(sentc.device)
        pred = self.hl + _pad_tensor(attn.sum(dim=2), bs) + self.label_bias  # shape (bs, n_lbs)
        
        if lbs_cf is not None: 
            lbs_cf = _pad_tensor(lbs_cf, bs)
            pred.add_(lbs_cf) 
        
        if self.y_range is not None: pred = sigmoid_range(pred, *self.y_range)
        return pred, attn, wgts, lbs_cf, lm_loss 

# %% ../../../nbs/02_text.models.core.ipynb 25
def get_xmltext_classifier(arch, vocab_sz, n_class, seq_len=72, config=None, drop_mult=1., pad_idx=1, max_len=72*20, y_range=None):
    "Create a text classifier from `arch` and its `config`, maybe `pretrained`"
    meta = _model_meta[arch]
    config = ifnone(config, meta['config_clas']).copy()
    for k in config.keys():
        if k.endswith('_p'): config[k] *= drop_mult
    n_hidden = config[meta['hid_name']]
    config.pop('output_p')
    init = config.pop('init') if 'init' in config else None
    encoder = SentenceEncoder(seq_len, arch(vocab_sz, **config), pad_idx=pad_idx, max_len=max_len)
    decoder = LabelAttentionClassifier(n_hidden, n_class, y_range=y_range)
    model = SequentialRNN(encoder, decoder)
    return model if init is None else model.apply(init)

# %% ../../../nbs/02_text.models.core.ipynb 26
import re
def _get_filter_params(m, but:re.Pattern):
    return [p for n, p in list(m.named_parameters()) if not but.match(n)]

# %% ../../../nbs/02_text.models.core.ipynb 27
def awd_lstm_xclas_split(model):
    "Split a RNN `model` in groups for differential learning rates."
    groups = [nn.Sequential(model[0].module.encoder, model[0].module.encoder_dp)]
    groups += [nn.Sequential(rnn, dp) for rnn, dp in zip(model[0].module.rnns, model[0].module.hidden_dps)]
    return L(groups).map(params) + [params(model[1].pay_attn.lm_decoder)] + [params(model[1].pay_attn.l2r)] + L(model[1]).map(partial(_get_filter_params, but=re.compile('^(pay_attn.l2r|pay_attn.lm_decoder)')))

# %% ../../../nbs/02_text.models.core.ipynb 29
def get_xmltext_classifier2(arch, vocab_sz, n_class, seq_len=72, config=None, drop_mult=1., pad_idx=1, max_len=72*20, y_range=None, running_decoder=True, 
                           plant=0.5, attn_init=(0, 0, 1), static_inattn=5, diff_inattn=30, lowshot=False, unfreeze_lm_decoder=False):
    "Create a text classifier from `arch` and its `config`, maybe `pretrained`"
    meta = _model_meta[arch]
    config = ifnone(config, meta['config_clas']).copy()
    for k in config.keys():
        if k.endswith('_p'): config[k] *= drop_mult
    n_hidden = config[meta['hid_name']]
    config.pop('output_p')
    init = config.pop('init') if 'init' in config else None
    decoder = LabelAttentionClassifier(n_hidden, n_class, plant=plant, attn_init=attn_init, y_range=y_range, static_inattn=static_inattn, diff_inattn=diff_inattn, lowshot=lowshot, unfreeze_lm_decoder=unfreeze_lm_decoder)
    encoder = AttentiveSentenceEncoder(seq_len, arch(vocab_sz, **config), decoder, pad_idx=pad_idx, max_len=max_len, running_decoder=running_decoder)
    model =  SequentialRNN(encoder, decoder)
    return model if init is None else model.apply(init)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/07_l2r.models.core.ipynb.

# %% auto 0
__all__ = ['L2R_DotProductBias', 'L2R_NN']

# %% ../../../nbs/07_l2r.models.core.ipynb 2
from fastai.torch_imports import *
from fastai.layers import *
from ...layers import *

# %% ../../../nbs/07_l2r.models.core.ipynb 6
class L2R_DotProductBias(nn.Module):
    def __init__(self, num_lbs, num_toks, num_factors, y_range=None):
        super().__init__()
        self.num_toks, self.num_lbs = num_toks+1, num_lbs+1 # +1 for the `padding_idx` 
        self.num_factors = num_factors
        self.token_factors = nn.Embedding(self.num_toks, num_factors, padding_idx=-1)
        self.token_bias = nn.Embedding(self.num_toks, 1, padding_idx=-1)
        self.label_factors = nn.Embedding(self.num_lbs, num_factors, padding_idx=-1)
        self.label_bias = nn.Embedding(self.num_lbs, 1, padding_idx=-1)
        self.y_range = y_range
        
    def forward(self, xb):
        # import pdb; pdb.set_trace()
        xb_toks = xb[:, :, :, 0].long() # xb[...,0] # shape (64, 2233, 64)
        xb_lbs = torch.unique(xb[:, :, :, 1], dim=-1).flatten(start_dim=1).long() # shape (64, 2233, )
        # To convert -1 which is the padding index to the last index:
        xb_toks, xb_lbs= xb_toks%(self.num_toks), xb_lbs%(self.num_lbs)
        
        toks_embs = self.token_factors(xb_toks) # shape (64, 2233, 64, 400)
        toks_shape = toks_embs.shape
        toks_embs = toks_embs.view(-1, *toks_shape[2:]) # shape (64*2233, 64, 400)

        lbs_embs = self.label_factors(xb_lbs) # shape (64, 2233, 400)
        lbs_shape = lbs_embs.shape
        lbs_embs = lbs_embs.view(-1, *lbs_shape[2:]).unsqueeze(dim=-1) # shape (64*2233, 400, 1)
        
        # res = torch.bmm(toks_embs, lbs_embs) # shape (64*2233, 64, 1)
        # res = res.view(toks_shape[0], toks_shape[1], *res.shape[1:]) + self.token_bias(xb_toks) + self.label_bias(xb_lbs).unsqueeze(2) # shape (65, 2233, 64, 1)

        bias = self.token_bias(xb_toks) + self.label_bias(xb_lbs).unsqueeze(2) # (bs, lbs_in_one_chunk, sl, 1)
        bias = bias.view(-1, *bias.shape[2:]) # (bs*lbs_in_one_chunk, sl, 1)
        res = torch.baddbmm(bias, toks_embs, lbs_embs) # toks_embs: (bs*lbs_in_one_chunk, sl, emb_sz), lbs_embs: (bs*lbs_in_one_chunk, emb_sz, 1), res : (bs*lbs_in_one_chunk, sl, 1)
        res = res.view(toks_shape[0], toks_shape[1], *res.shape[1:]) # res (bs, lbs_in_one_chunk, sl, 1)
        
        return sigmoid_range(res, *self.y_range) if self.y_range is not None else res
        # return res

# %% ../../../nbs/07_l2r.models.core.ipynb 8
class L2R_NN(nn.Module):
    def __init__(self, num_lbs, 
                 num_toks, 
                 num_factors, 
                 layers,
                 ps=None,
                 use_bn=True,
                 bn_final=False,
                 lin_first=True,
                 embed_p=0.0,
                 y_range=None):
        super().__init__()
        self.num_toks, self.num_lbs = num_toks+1, num_lbs+1 # +1 for the `padding_idx` 
        self.num_factors, self.embed_p, self.ps = num_factors, embed_p, ps
        self.token_factors = nn.Embedding(self.num_toks, num_factors, padding_idx=-1)
        self.label_factors = nn.Embedding(self.num_lbs, num_factors, padding_idx=-1)
        self.emb_drop = nn.Dropout(embed_p)
        self.y_range = y_range
        if ps is None: self.ps = [0. for _ in range(len(layers))]
        sizes = [self.token_factors.embedding_dim] + layers + [1]
        actns = [nn.ReLU(inplace=True) for _ in range(len(sizes)-2)] + [None]
        _layers = [LinBnFlatDrop(sizes[i], sizes[i+1], bn=use_bn and (i!=len(actns)-1 or bn_final), p=p, act=a, lin_first=lin_first)
                   for i, (p,a) in enumerate(zip(self.ps+[0.],actns))]
        self.layers= nn.Sequential(*_layers)
        # self.layers = nn.Sequential(
        #     nn.Linear(num_factors, n_act),
        #     nn.ReLU(),
        #     nn.Linear(n_act, 1),
        #     nn.Dropout(self.ps),
        # )
        
    # def __str__(self): return super().__repr__() + f"\n {self.n_act = }, {self.embed_p = }"
    # __repr__ = __str__

    
    def forward(self, xb):
        # import pdb; pdb.set_trace()
        xb_toks = xb[:, :, :, 0].long() # xb[...,0] # shape (64, 2233, 64)
        xb_lbs = torch.unique(xb[:, :, :, 1], dim=-1).flatten(start_dim=1).long() # shape (64, 2233, )
        # To convert -1 which is the padding index to the last index:
        xb_toks, xb_lbs= xb_toks%(self.num_toks), xb_lbs%(self.num_lbs)
        
        toks_embs = self.token_factors(xb_toks) # shape (64, 2233, 64, 200)

        lbs_embs = self.label_factors(xb_lbs) # shape (64, 2233, 200)
        lbs_embs = lbs_embs.unsqueeze(2) # shape (64, 2233, 1, 200)
        lbs_embs = lbs_embs.expand(-1, -1, xb.shape[2], -1)
        
        # embs = torch.cat((toks_embs, lbs_embs), dim=-1) # shape (64, 2233, 64, 400)
        embs = toks_embs + lbs_embs
        embs = self.emb_drop(embs)
        res = self.layers(embs)
        
        return sigmoid_range(res, *self.y_range) if self.y_range is not None else res
        # return res



from fastai.imports import *

import numpy as np
import pandas as pd
from scipy.stats import *
import matplotlib.pyplot as plt
import seaborn as sns
import itertools
from pathlib import Path
from enum import Enum
from collections import OrderedDict,defaultdict,Counter,namedtuple
from contextlib import suppress
import subprocess
import json
import ast
from tqdm.notebook import tqdm, trange
import tempfile
import os
from icecream import ic
from IPython.display import clear_output
import pdb
from fastprogress.fastprogress import progress_bar,master_bar
from fastcore.all import *



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils.ipynb.

# %% auto 0
__all__ = ['namestr', 'list_files', 'make_paths', 'plot_hist', 'plot_reduction', 'test_eqs', 'validate', 'cudamem',
           'get_description', 'is_colab']

# %% ../nbs/00_utils.ipynb 2
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from fastcore.all import *
from .imports import *
from .torch_imports import *
from .fastai_imports import *

# %% ../nbs/00_utils.ipynb 5
def namestr(obj, namespace=None):
    "Returns the name of the object `obj` passed"
    return [name for name in namespace if namespace[name] is obj]

# %% ../nbs/00_utils.ipynb 8
def list_files(startpath):
    """ [simulates the linux tree cmd] 
    (https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python)
    """ 
    
    if isinstance(startpath, Path): startpath = str(startpath) 
    for root, dirs, files in os.walk(startpath):
        level = root.replace(startpath, '').count(os.sep)
        indent = ' ' * 4 * (level)
        print('{}{}/'.format(indent, os.path.basename(root)))
        subindent = ' ' * 4 * (level + 1)
        for f in files:
            print('{}{}'.format(subindent, f))

# %% ../nbs/00_utils.ipynb 9
def make_paths(path, prefix=None):
    """
    with `path` as basedir, makes data and models dir and 
    returns a dictionary of relevant pathlib objects
    """
    path_data = path/'data'
    path_model = path/'models'

    path_model.mkdir(exist_ok=True)
    path_data.mkdir(exist_ok=True)
    (path_model/'collab').mkdir(exist_ok=True)

    data = path_data/(prefix+'.csv')
    dls_lm_path, dls_lm_r_path = path_model/f"{prefix}_dls_lm.pkl", path_model/f"{prefix}_dls_lm_r.pkl"
    dls_lm_vocab_path, dls_lm_vocab_r_path = path_model/f"{prefix}_dls_lm_vocab.pkl", path_model/f"{prefix}_dls_lm_vocab_r.pkl"
    lm_path, lm_r_path = path_model/f"{prefix}_lm.pth", path_model/f"{prefix}_lm_r.pth"
    lm_finetuned_path, lm_finetuned_r_path = path_model/f"{prefix}_lm_finetuned.pth", path_model/f"{prefix}_lm_finetuned_r.pth"
    dsets_clas_path, dsets_clas_r_path = path_model/f"{prefix}_dset_clas.pkl", path_model/f"{prefix}_dset_clas_r.pkl"
    dls_clas_path, dls_clas_r_path = path_model/f"{prefix}_dls_clas.pkl", path_model/f"{prefix}_dls_clas_r.pkl"
    clas_path, clas_r_path = path_model/f"{prefix}_clas.pth", path_model/f"{prefix}_clas_r.pth"
    collab_bootst_path = path_model/f"{prefix}_tok_lbl_info.pkl"
    collab_data_path = path_data/f"{prefix}_tok_lbl.ft"
    collab_tok_path = path_data/f"{prefix}_tok.ft"
    collab_lbl_path = path_data/f"{prefix}_lbl.ft"
    dls_collab_path = path_model/f"{prefix}_dls_collab.pkl"
    dls_learn_rank_path = path_model/f"{prefix}_dls_learn_rank.pkl"
    collab_path = path_model/'collab'/f"{prefix}_collab.pth"
    plist = [path, path_data, path_model, 
             data, 
             dls_lm_path, dls_lm_r_path,
             dls_lm_vocab_path, dls_lm_vocab_r_path,
             lm_path, lm_r_path,
             lm_finetuned_path, lm_finetuned_r_path,
             dsets_clas_path, dsets_clas_r_path,
             dls_clas_path, dls_clas_r_path,
             clas_path, clas_r_path,
             collab_bootst_path,
             collab_data_path,
             collab_tok_path,
             collab_lbl_path,
             dls_collab_path,
             dls_learn_rank_path,
             collab_path]
    pdir = {}
    for o in plist:  pdir[namestr(o, locals())[0]] = o
    return pdir

# %% ../nbs/00_utils.ipynb 11
def plot_hist(data, x_label=None, y_label=None, title="Histogram"):
    n, bins, pathches = plt.hist(data)
    plt.grid(axis='y', color='b')
    # plt.yscale('log')
    if x_label is not None: plt.xlabel(x_label)
    if y_label is not None: plt.ylabel(y_label)
    maxfreq = n.max()
    plt.ylim(ymax=np.ceil(maxfreq / 10) * 10 if maxfreq % 10 else maxfreq + 10)
    plt.title(title);

# %% ../nbs/00_utils.ipynb 12
def plot_reduction(X, tSNE=True, n_comps=None, perplexity=30, figsize=(6,4)):
    """
    PCA on X and plots the first two principal components, returns the decomposition 
    and the explained variances for each directions,
    if `tSNE` then does a tSNE after PCA.
    """
    reduction = "tSNE" if tSNE else "PCA"
    pca = PCA(n_components=n_comps, svd_solver="full")
    X_red = pca.fit_transform(X)
    if tSNE:
        tsne = TSNE(n_components=2, perplexity=perplexity)
        X_red = tsne.fit_transform(X_red[:, :50])
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1,1,1)
    plt.scatter(X_red[:, 0], X_red[:, 1], marker='x')
    ax.set_xlabel("1st component")
    ax.set_ylabel("2nd component")
    ax.set_title(f"{reduction} Decomposition")
    plt.show()
    return X_red, pca.explained_variance_ratio_

# %% ../nbs/00_utils.ipynb 13
def test_eqs(*args):
    for i in range(len(args)-1):
        test_eq(args[i], args[i+1])

# %% ../nbs/00_utils.ipynb 15
from fastai.callback.tracker import SaveModelCallback

# %% ../nbs/00_utils.ipynb 16
def validate(learner, cb=SaveModelCallback, **kwargs):
    "validates a `learner` within a context manager after temporarily removing `cb` if it exists"
    save_cb_idx = learner.cbs.argfirst(lambda o: isinstance(o, cb))
    if save_cb_idx is None:
        print(learner.validate(**kwargs))
        return
    print(f'best so far = {learner.cbs[save_cb_idx].best}')
    with learner.removed_cbs(learner.cbs[save_cb_idx]):
        vals = learner.validate(**kwargs)
        names = learner.recorder._valid_mets.map(Self.name())
        print('\n'.join([f"{n} = {v}" for n,v in zip(names,vals)]))
    save_cb_idx = learner.cbs.argfirst(lambda o: isinstance(o, cb))
    print(f'best so far = {learner.cbs[save_cb_idx].best}')
    return vals

# %% ../nbs/00_utils.ipynb 17
def cudamem(device=default_device()):
    using = torch.cuda.max_memory_reserved(device)/1024**3
    total = torch.cuda.get_device_properties(device).total_memory/1024**3
    print(f"GPU: {torch.cuda.get_device_name(default_device())}")
    print(f"You are using {using} GB")
    print(f"Total GPU memory = {total} GB")

# %% ../nbs/00_utils.ipynb 18
import requests
from bs4 import BeautifulSoup
def get_description(codes):
    "descriptions of ICD10 codes"
    # Initialize an empty dictionary to store code descriptions
    icd10_descriptions = {}
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"
    }
    with requests.Session() as session:
        for code in codes:
            url_cm = f"https://icd10coded.com/cm/{code}"
            url_pcs = f"https://icd10coded.com/pcs/{code}"
            
            for url in (url_pcs, url_cm):   
                with session.get(url, headers=headers, timeout=5) as response:
                    if response.status_code == 200:
                        soup = BeautifulSoup(response.content, 'html.parser')
                        description_element = soup.find('span', {'class': 'lead'})
                        if description_element:
                            description = description_element.text.strip()
                            icd10_descriptions[code] = description
                            break
                        else:
                            icd10_descriptions[code] = "Description not found"
                    else:
                        # print(f"An error occurred.")
                        icd10_descriptions[code] = "Code not found"
    return icd10_descriptions

# %% ../nbs/00_utils.ipynb 19
def is_colab():
    # Check if the 'google.colab' module is available
    try:
        import google.colab
        IN_COLAB = True
    except ImportError:
        IN_COLAB = False

    # Check for specific environment variables typically set in Colab
    if 'COLAB_GPU' in os.environ:
        IN_COLAB = True

    if IN_COLAB:
        print("Running in Google Colab")
    else:
        print("Not running in Google Colab")

    if IN_COLAB:
        try:
            subprocess.check_call(["pip", "install", 'GitPython'])
            print(f"Successfully installed GitPython")
            from git import Repo
            repo = Repo.clone_from('https://github.com/debjyotiSRoy/xcube.git', Path.cwd().parent/'xcube')
            repo.git.checkout('plant')
        except subprocess.CalledProcessError as e:
            print(f"Failed to install GitPython: {e}")
    return IN_COLAB



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/12_data.external.ipynb.

# %% ../../nbs/12_data.external.ipynb 2
from __future__ import annotations
from fastdownload import FastDownload
from functools import lru_cache
from ..imports import *
import xcube.data

# %% auto 0
__all__ = ['xcube_cfg', 'xcube_path', 'XURLs', 'untar_xxx']

# %% ../../nbs/12_data.external.ipynb 7
@lru_cache(maxsize=None)
def xcube_cfg() -> Config: # Config that contains default download paths for `data`, `model`, `storage` and `archive`
    "`Config` object for xcube's `config.ini`"
    return Config(Path(os.getenv('XCUBE_HOME', '~/.xcube')), 'config.ini', create=dict(
        data = 'data', archive = 'archive', storage = 'tmp', model = 'models'))

# %% ../../nbs/12_data.external.ipynb 10
def xcube_path(folder:str) -> Path: 
    "Local path to `folder` in `Config`"
    return xcube_cfg().path(folder)

# %% ../../nbs/12_data.external.ipynb 13
class XURLs():
    "Global cosntants for datasets and model URLs."
    LOCAL_PATH = Path.cwd()
    S3 = 'https://xcubebucket.s3.us-east-2.amazonaws.com/'
    
    #main datasets
    MIMIC3 = f'{S3}mimic3/mimic3.tgz'
    MIMIC3_DEMO = f'{S3}mimic3/mimic3_demo.tgz'
    MIMIC3_L2R = f'{S3}mimic3/mimic3_l2r/mimic3_l2r.tgz'
    MIMIC4 = f'{S3}mimic4/mimic4.tgz'
    MIMIC4_L2R = f'{S3}mimic4/mimic4_l2r/mimic4_l2r.tgz'
    
    def path(
        url:str='.', # File to download
        c_key:str='archive' # Key in `Config` where to save URL
    ) -> Path:
        "Local path where to download based on `c_key`"
        fname = url.split('/')[-1]
        local_path = XURLs.LOCAL_PATH/('models' if c_key=='model' else 'data')/fname
        if local_path.exists(): return local_path
        return xcube_path(c_key)/fname

# %% ../../nbs/12_data.external.ipynb 18
def untar_xxx(
    url:str, # File to download
    archive:Path=None, # Optional override for `Config`'s `archive` key
    data:Path=None, # Optional override for `Config`'s `data` key
    c_key:str='data', # Key in `Config` where to extract file
    force_download:bool=False, # Setting to `True` will overwrite any existing copy of data
    base:str='~/.xcube' # Directory containing config file and base of relative paths
) -> Path: # Path to extracted file(s)
    "Download `url` using `FastDownload.get`"
    d = FastDownload(xcube_cfg(), module=xcube.data, archive=archive, data=data, base=base)
    return d.get(url, force=force_download, extract_key=c_key)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/13_l2r.data.info_gain_old.ipynb.

# %% auto 0
__all__ = ['BatchLbsChunkify', 'MutualInfoGain']

# %% ../../../nbs/13_l2r.data.info_gain_old.ipynb 2
from fastcore.basics import *
from fastai.torch_core import *
from fastai.data.core import *
from fastai.data.transforms import *
from fastai.text.core import *
from fastai.text.data import *
from ...imports import *
from ...torch_imports import *

# %% ../../../nbs/13_l2r.data.info_gain_old.ipynb 6
class BatchLbsChunkify(ItemTransform):
    order = 100
    def __init__(self, chnk_st, chnk_end): store_attr('chnk_st,chnk_end')
    def encodes(self, x): 
        return (x[0], x[1][:, self.chnk_st:self.chnk_end])

# %% ../../../nbs/13_l2r.data.info_gain_old.ipynb 7
class MutualInfoGain:
    def __init__(self, df, bs=8, chnk_sz=200, device=None): store_attr('df,bs,chnk_sz,device')
    
    def onehotify(self):
        x_tfms = [Tokenizer.from_df('text', n_workers=num_cpus()), attrgetter("text"), Numericalize(), OneHotEncode()]
        y_tfms = [ColReader('labels', label_delim=';'), MultiCategorize(), OneHotEncode()]
        tfms = [x_tfms, y_tfms]
        self.dsets = Datasets(self.df, tfms=[x_tfms, y_tfms], )
        self.toksize, self.lblsize = self.dsets.vocab.map(len)
        return self.dsets
        
    def lbs_chunked(self):
        lbs = self.dsets.vocab[1]
        self.dls = []
        for chnk_st in range(0, len(lbs), self.chnk_sz):
            self.dls.append(TfmdDL(self.dsets, bs=self.bs, 
                              after_batch=[BatchLbsChunkify(chnk_st, min(chnk_st+self.chnk_sz, len(lbs)))], 
                              device=default_device() if self.device is None else self.device))
        return self.dls
    
    def _mutual_info_gain(self, dl):
        """
        Computes [mutual information gain](https://en.wikipedia.org/wiki/Mutual_information) for each token label pair
        `dl` is (bag-of-words text, one-hot encoded targets)
        """
        xb, yb = dl.one_batch() 
        toksize, lblsize = xb.size(1), yb.size(1)
        p_TL = torch.zeros(toksize, lblsize, 4, dtype=torch.float, device=default_device())
        eps = p_TL.new_empty(1).fill_(1e-8)
        for x,y in dl:
            test_eq(x.shape, (dl.bs, toksize)); test_eq(y.shape, (dl.bs, lblsize))
            t = x.unsqueeze(-1).expand(-1, -1, lblsize) ; test_eq(t.shape, (dl.bs, toksize, lblsize))
            l = y.unsqueeze(1).expand(-1, toksize, -1) ; test_eq(l.shape, (dl.bs, toksize, lblsize))
            tl = torch.stack((t,l), dim=-1) ; test_eq(tl.shape, (dl.bs, toksize, lblsize, 2)) 
            p_TL_tt = tl[...,0].logical_and(tl[...,1]) ; test_eq(p_TL_tt.shape, (dl.bs, toksize, lblsize)) 
            p_TL_tf = tl[...,0].logical_and(tl[...,1].logical_not()) ; test_eq(p_TL_tf.shape, (dl.bs, toksize, lblsize)) 
            p_TL_ft = tl[...,0].logical_not().logical_and(tl[...,1]) ; test_eq(p_TL_ft.shape, (dl.bs, toksize, lblsize))
            p_TL_ff = tl[...,0].logical_not().logical_and(tl[...,1].logical_not()) ; test_eq(p_TL_ff.shape, (dl.bs, toksize, lblsize)) 
            p_TL = p_TL + torch.stack((p_TL_tt, p_TL_tf, p_TL_ft, p_TL_ff), dim=-1).float().sum(dim=0)
        p_TL = p_TL / tensor(len(self.dsets)).float()
        p_TL = p_TL.view(toksize, lblsize, 2, 2) ; test_eq(p_TL.shape, (toksize, lblsize, 2, 2))# last axis: lbl axis, 2nd last axis: token axis
        return p_TL
    
    def joint_pmf(self):
        self.p_TL_full = [] 
        for dl in progress_bar(self.dls):
            p_TL = self._mutual_info_gain(dl)
            self.p_TL_full.append(p_TL)
            del p_TL; #del p_T; del p_L; del p_TxL; del I_TL; torch.cuda.empty_cache()
        self.p_TL_full = torch.cat(self.p_TL_full, dim=1); test_eq(self.p_TL_full.shape, (self.toksize, self.lblsize, 2, 2))
        return self.p_TL_full
    
    def compute(self):
        eps = self.p_TL_full.new_empty(1).fill_(1e-15)
        toksize, lblsize = self.p_TL_full.size(0), self.p_TL_full.size(1)
        p_T = self.p_TL_full[:,0].sum(-1, keepdim=True); test_eq(p_T.shape, (toksize, 2, 1))# 0 because we can pick any label and apply total prob law
        p_L = self.p_TL_full[0,:].sum(-2, keepdim=True); test_eq(p_L.shape, (lblsize, 1, 2)) # 0 becuase we can pick any token and apply total prob law
        p_TxL = self.p_TL_full.sum(-1, keepdim=True) @ self.p_TL_full.sum(-2, keepdim=True); test_eq(p_TxL.shape, (toksize, lblsize, 2, 2))
        H_T = -(p_T * torch.log(p_T+eps)).sum(-2).squeeze(); test_eq(H_T.shape, [toksize])
        H_L = -(p_L * torch.log(p_L+eps)).sum(-1).squeeze(); test_eq(H_L.shape, [lblsize])
        I_TL = (self.p_TL_full * torch.log((self.p_TL_full + eps)/(p_TxL + eps))).flatten(start_dim=-2).sum(-1); test_eq(I_TL.shape, (toksize, lblsize))
        return p_T, p_L, p_TxL, H_T, H_L, I_TL

# %% ../../../nbs/13_l2r.data.info_gain_old.ipynb 8
@patch
def lbs_frqs(self:MutualInfoGain):
    f = ColReader('labels', label_delim=';')
    frqs = Counter()
    for o in self.df.itertuples(): frqs.update(f(o))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/11_l2r.callbacks.ipynb.

# %% auto 0
__all__ = ['TrainEval', 'AvgSmoothMetric', 'TrackResults', 'Monitor', 'XParamScheduler', 'XLRFinder', 'ProgressBarCallback',
           'SaveCallBack', 'InterruptCallback', 'CSVLog', 'AMPMode', 'TestCallback', 'MemoryErrorTest',
           'MixedPrecision', 'GradientClip']

# %% ../../nbs/11_l2r.callbacks.ipynb 2
from fastai.torch_imports import *
from fastai.torch_core import *
from fastai.callback.core import *
from fastcore.all import *
from ..imports import *
from ..metrics import *

# %% ../../nbs/11_l2r.callbacks.ipynb 6
class TrainEval(Callback):
    "`Callback` that tracks the number of iterations done and properly sets training/eval mode"
    order, run_valid = -10, False
    
    def before_fit(self):
        "Set the iter and epoch counters to 0"
        self.learn.epoch = 0
        self.learn.train_iter,self.learn.pct_train = 0,0.
        
    def after_batch(self):
        "Update the iter counter (in training mode)"
        if not self.model.training: return
        self.learn.pct_train += 1./(self.n_iter*self.n_epochs)
        self.learn.train_iter += 1
        
    def before_train(self):
        "Set the model to training mode"
        self.learn.pct_train=self.epoch/self.n_epochs
        self.model.train()
        self.learn.training=True
        
    def before_validate(self):
        "Set the model to validation mode"
        self.model.eval()
        self.learn.training=False

# %% ../../nbs/11_l2r.callbacks.ipynb 8
from fastai.learner import Metric

# %% ../../nbs/11_l2r.callbacks.ipynb 9
class AvgSmoothMetric(Metric):
    "Smooth average of the losses (exponentially weighted with `beta`)"
    def __init__(self, beta=0.98): self.beta = beta
    def reset(self):               self.count,self.val = 0,tensor(0.)
    def accumulate(self, learn):
        self.count += 1
        self.val = torch.lerp(to_detach(learn.moi.mean()), self.val, self.beta)
    @property
    def value(self): return self.val/(1-self.beta**self.count)

# %% ../../nbs/11_l2r.callbacks.ipynb 11
from fastprogress.core import format_time  
import time

# %% ../../nbs/11_l2r.callbacks.ipynb 12
class TrackResults(Callback):
    def __init__(self, train_metrics=False, beta=0.98): 
        store_attr()
        self.names = ['loss', 'ndcg', 'ndcg_at_6', 'acc']
        self.metric_names = L(self.names).map(lambda o: 'val_'+o)
        if train_metrics: self.metric_names = L(self.names).map(lambda o: 'train_'+o) + self.metric_names
        self.metric_names.append('time')
        self.metric_names = L('epoch') + self.metric_names
        self.smooth_moi = AvgSmoothMetric(beta=beta) # to maintain weighted avg of metric of interest 
    
    def before_fit(self): 
        self.lrs, self.mois, self.smooth_mois, self.losses_full, self.grads_full, self.metrics_full = [], [], [], [], defaultdict(list), defaultdict(list) 
        self.ioi = self.names.index(self.learn.save_call_back.monitor
                                  if hasattr(self.learn, 'save_call_back') else 'loss')
        self.smooth_moi.reset()
    
    def before_train(self): self._initialise_metrics()
    
    def before_validate(self): self._initialise_metrics()
        
    def after_train(self):
        self.losses_full.extend(self.losses)
        log = self._compute_epoch_mean()
        if self.train_metrics:
            self.metrics_full['trn'].append(log)
        # print(self.epoch, self.model.training, *log)
        self.log+=log
                
    def after_validate(self):
        log = self._compute_epoch_mean()
        if hasattr(self, 'metrics_full'):
            self.metrics_full['val'].append(log)
        # print(self.epoch if hasattr(self, 'epoch') else 0, self.model.training, *log)
        if getattr(self, 'log', False): self.log+=log

    def before_epoch(self):
        self.start_epoch = time.time()
        self.log = L(getattr(self, 'epoch', 0))

    def after_epoch(self): 
        self.log.append(format_time(time.time() - self.start_epoch))
        # import pdb; pdb.set_trace()
        self.logger(self.log)
            
    def _compute_epoch_mean(self):
        _li = [self.losses, self.ndcgs, self.ndcgs_at_6, self.accs]
        _li = [torch.stack(o) if o else torch.Tensor() for o in _li] 
        [self.losses, self.ndcgs, self.ndcgs_at_6, self.accs] = _li
        log = [round(o.mean().item(), 4) if o.sum() else "NA" for o in _li]
        return log
    
    def _initialise_metrics(self): self.losses, self.ndcgs, self.ndcgs_at_6, self.accs = [], [], [], []

    
    def after_batch(self):
        with torch.no_grad():
            loss = self.loss_func(self.preds, self.xb)
            if loss.mean().isnan(): 
                import pdb; pdb.set_trace()
            self.losses.append(loss.mean())
            if self.model.training:
                if self.train_metrics: 
                    self._compute_metrics()
                    # grab moi
                    self.learn.moi = getattr(self, "losses ndcgs ndcgs_at_6 accs".split()[self.ioi])[-1] # -1 is to get the value for the current batch
                    # do weighted avg
                    self.smooth_moi.accumulate(self.learn)
                    # record `learn` attribute
                    self.learn.smooth_moi = self.smooth_moi.value
                    # record moi attribute in `self`
                    self.mois.append(self.learn.moi)
                    # record smooth_moi attribute in `self`
                    self.smooth_mois.append(self.learn.smooth_moi)
                    # record lr in `self`
                    self.lrs.append(self.opt.hypers[-1]['lr'])
            else: self._compute_metrics()
                        
    def _compute_metrics(self):
        *_, _ndcg, _ndcg_at_k = ndcg(self.preds, self.xb, k=6)
        self.ndcgs.append(_ndcg.mean())
        self.ndcgs_at_6.append(_ndcg_at_k.mean())
        acc = accuracy(self.xb, self.model).mean()
        self.accs.append(acc.mean())
        
    def after_backward(self):
        for name,param in self.model.named_parameters():
            grad = param.grad.data.detach().clone()
            self.grads_full[name].append(grad)

# %% ../../nbs/11_l2r.callbacks.ipynb 13
class Monitor(Callback):
    order = 60
    def __init__(self, monitor='ndcg_at_6', comp=None, min_delta=0., reset_on_fit=False):
        if comp is None: comp = np.greater
        if comp == np.less: min_delta *= -1
        # store_attr()
        self.monitor,self.comp,self.min_delta,self.reset_on_fit,self.best= monitor,comp,min_delta,reset_on_fit,None
       
    def before_fit(self):
        if self.reset_on_fit or self.best is None: self.best = float('inf') if self.comp == np.less else -float('inf')
        assert self.monitor in self.track_results.names
        self.idx = list(self.track_results.names).index(self.monitor)
        
    def after_epoch(self):
        if self.track_results.metrics_full.get('val') is None: return 
        val = self.track_results.metrics_full.get('val')[-1][self.idx]
        if self.comp(val - self.min_delta if not isinstance(val, str) else self.best, self.best): self.best, self.new_best, = val, True
        else: self.new_best = False

# %% ../../nbs/11_l2r.callbacks.ipynb 15
@docs
class XParamScheduler(Callback):
    "Schedule hyper-parameters according to `scheds`"
    order,run_valid = 60,False

    def __init__(self, scheds): self.scheds = scheds
    def before_fit(self): self.hps = {p:[] for p in self.scheds.keys()}
    def before_batch(self): 
        if not self.model.training: return
        self._update_val(self.pct_train)

    def _update_val(self, pct):
        for n,f in self.scheds.items(): self.opt.set_hyper(n, f(pct))

    def after_batch(self):
        if not self.model.training: return
        for p in self.scheds.keys(): self.hps[p].append(self.opt.hypers[-1][p])

    def after_fit(self):
        if hasattr(self.learn, 'track_results') and hasattr(self, 'hps'): self.track_results.hps = self.hps

    _docs = {"before_fit": "Initialize container for hyper-parameters",
             "before_batch": "Set the proper hyper-parameters in the optimizer",
             "after_batch": "Record hyper-parameters of this batch",
             "after_fit": "Save the hyper-parameters in the track_results if there is one"}

# %% ../../nbs/11_l2r.callbacks.ipynb 21
from .learner import L2RLearner
from fastai.callback.schedule import *

# %% ../../nbs/11_l2r.callbacks.ipynb 22
@patch
def fit_one_cycle(self:L2RLearner, n_epoch, lr_max=None, div=25., div_final=1e5, pct_start=0.25, moms=None, cbs=None, start_epoch=0):
    "Fit `self.model` for `n_epoch` using the 1cycle policy."
    self.lr_max = lr_max
    self.opt = getattr(self, 'opt', None)
    if self.opt is None: self.create_opt()
    self.opt.set_hyper('lr', self.lr if lr_max is None else lr_max)
    lr_max = np.array([h['lr'] for h in self.opt.hypers])
    scheds = {'lr': combined_cos(pct_start, lr_max/div, lr_max, lr_max/div_final),
              'mom': combined_cos(pct_start, *(self.moms if moms is None else moms))}
    # self.add_cb(XParamScheduler(scheds)) 
    # pdb.set_trace()
    self.fit(n_epoch, cbs=XParamScheduler(scheds)+L(cbs), start_epoch=start_epoch)

# %% ../../nbs/11_l2r.callbacks.ipynb 24
@patch
def plot_sched(self:TrackResults, keys=None, figsize=None):
    keys = self.hps.keys() if keys is None else L(keys)
    rows,cols = (len(keys)+1)//2, min(2, len(keys))
    figsize = figsize or (6*cols,4*rows)
    _, axs = plt.subplots(rows, cols, figsize=figsize)
    axs = axs.flatten() if len(keys) > 1 else L(axs)
    for p,ax in zip(keys, axs):
        ax.plot(self.hps[p])
        ax.set_ylabel(p)

# %% ../../nbs/11_l2r.callbacks.ipynb 25
@patch
def fit_sgdr(self:L2RLearner, n_cycles, cycle_len, lr_max=None, cycle_mult=2, cbs=None, reset_opt=False, wd=None,
             start_epoch=0):
    "Fit `self.model` for `n_cycles` of `cycle_len` using SGDR."
    self.lr_max = lr_max
    self.opt = getattr(self, 'opt', None)
    if self.opt is None: self.create_opt()
    self.opt.set_hyper('lr', self.lr if lr_max is None else lr_max)
    lr_max = np.array([h['lr'] for h in self.opt.hypers])
    n_epoch = cycle_len * (cycle_mult**n_cycles-1)//(cycle_mult-1)
    pcts = [cycle_len * cycle_mult**i / n_epoch for i in range(n_cycles)]
    scheds = [SchedCos(lr_max, 0) for _ in range(n_cycles)]
    scheds = {'lr': combine_scheds(pcts, scheds)}
    self.fit(n_epoch, cbs=ParamScheduler(scheds)+L(cbs), reset_opt=reset_opt, wd=wd, start_epoch=start_epoch)

# %% ../../nbs/11_l2r.callbacks.ipynb 28
@docs
class XLRFinder(XParamScheduler):
    "Training with exponentially growing learning rate"
    run_after=TrackResults
    def __init__(self, start_lr=1e-7, end_lr=10, num_it=100, stop_div=True):
        if num_it < 6: num_it = 6
        self.scheds = {'lr': [SchedExp(s, e) for (s,e) in zip(start_lr,end_lr)
                             ] if is_listy(start_lr) else SchedExp(start_lr, end_lr)}
        self.num_it,self.stop_div = num_it,stop_div

    def before_fit(self):
        if hasattr(self, 'track_results'):
            moi = self.track_results.names[self.track_results.ioi]
            print(f'Smoothing {moi}')
        super().before_fit()
        path = self.path
        path.mkdir(parents=True, exist_ok=True)
        self.tmp_d = tempfile.TemporaryDirectory(dir=path)
        self.tmp_p = Path(self.tmp_d.name).stem
        self.learn.save(f'{self.tmp_p}/_tmp')
        self.best_moi = -float('inf')

    def before_batch(self): self._update_val(self.train_iter/self.num_it)

    def after_batch(self):
        super().after_batch()
        if self.smooth_moi > self.best_moi: self.best_moi = self.smooth_moi
        if 4*self.smooth_moi < self.best_moi and self.stop_div: raise CancelFitException()
        if self.train_iter >= self.num_it: raise CancelFitException()

    def before_validate(self): raise CancelValidException()

    def after_fit(self):
        self.learn.opt.zero_grad() # Needed before detaching the optimizer for future fits
        tmp_f = self.path/self.tmp_p/'_tmp.pth'
        if tmp_f.exists():
            self.learn.load(f'{self.tmp_p}/_tmp', with_opt=True)
            self.tmp_d.cleanup()
            
    def after_cancel_validate(self): pass

    _docs = {"before_fit": "Initialize container for hyper-parameters and save the model",
             "before_batch": "Set the proper hyper-parameters in the optimizer",
             "after_batch": "Record hyper-parameters of this batch and potentially stop training",
             "after_fit": "Save the hyper-parameters in the recorder if there is one and load the original model",
             "before_validate": "Skip the validation part of training",
             "after_cancel_validate": "pass `CancelValidException`",
             "run_after":""}

# %% ../../nbs/11_l2r.callbacks.ipynb 29
@patch
def plot_xlr_find(self:TrackResults, skip_end=5, return_fig=True, suggestions=None, nms=None, **kwargs):
    "Plot the result of an LR Finder test (won't work if you didn't do `learn.xlr_find()` before)"
    lrs  = self.lrs  if skip_end==0 else self.lrs[:-skip_end]
    mois = L(self.mois if skip_end==0 else self.mois[:-skip_end]).map(Tensor.cpu)
    smooth_mois = L(self.smooth_mois if skip_end==0 else self.smooth_mois[:-skip_end]).map(Tensor.cpu)
    fig, (ax1, ax2) = plt.subplots(1,2)
    ax1.plot(lrs, smooth_mois)
    ax1.set_ylabel("Smoothened MOI")
    ax1.set_xlabel("Learning Rate")
    ax1.set_xscale('log')
    if suggestions:
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color'][1:]
        for (val, idx), nm, color in zip(suggestions, nms, colors):
            ax1.plot(val, idx, 'o', label=nm, c=color)
        ax1.legend(loc='best')
    ax2.plot(range(len(mois)), mois, label='MOI')
    ax2.plot(range(len(smooth_mois)), smooth_mois, label='SMOI')
    ax2.legend(loc='best')


# %% ../../nbs/11_l2r.callbacks.ipynb 30
@patch
def xrl_find(self:L2RLearner, start_lr=1e-5, end_lr=1e-1, num_it=400, stop_div=True, show_plot=True, suggest_funcs=(valley,)):
    "Launch a mock training to find a good learning rate and return suggestions based on `suggest_funcs` as a named tuple"
    n_epochs = num_it//len(self.dls.train) + 1
    cb=XLRFinder(start_lr=start_lr, end_lr=end_lr, num_it=num_it, stop_div=stop_div)
    self.fit(n_epochs, cbs=cb)
    if suggest_funcs is not None:
        lrs, smooth_mois = tensor(self.track_results.lrs[num_it//10:-5]),tensor(self.track_results.smooth_mois[num_it//10:-5])
        nan_idxs = torch.nonzero(torch.isnan(smooth_mois.view(-1)))
        if len(nan_idxs):
            drop_idx = min(nan_idxs)
            lrs = lrs[:drop_idx]
            smooth_mois = smooth_mois[:drop_idx]
        _suggestions, nms = [], []
        for func in tuplify(suggest_funcs):
            nms.append(func.__name__ if not isinstance(func, partial) else func.func.__name__) # deal with partials
            _suggestions.append(func(lrs, smooth_mois, num_it))
            
        # pdb.set_trace()
        SuggestedLRs = collections.namedtuple('SuggestedLRs', nms)
        lrs, pnts = [], []
        for lr, pnt in _suggestions:
            lrs.append(lr)
            pnts.append(pnt)   
        if show_plot: self.track_results.plot_xlr_find(suggestions=pnts, nms=nms)
        return SuggestedLRs(*lrs)
    
    elif show_plot: self.track_results.plot_xlr_find()
        

# %% ../../nbs/11_l2r.callbacks.ipynb 32
class ProgressBarCallback(Callback):
    order = 70
    
    def before_fit(self):
        self.mbar = master_bar(range(self.n_epochs))
        if self.learn.logger != noop:
            self.old_logger,self.learn.logger = self.logger,self._write_stats
            # metric_names = L(self.track_results.names).map(lambda o: 'train_'+o) + L(self.track_results.names).map(lambda o: 'val_'+o)
            # metric_names.append('time')
            # metric_names = L('epoch') + metric_names
            self._write_stats(self.track_results.metric_names)
        
    def before_epoch(self):
        if getattr(self, 'mbar', False): self.mbar.update(self.epoch)
        
    def before_train(self): self._launch_pbar()
    
    def before_validate(self): self._launch_pbar()
        
    def _launch_pbar(self):
        self.pbar = progress_bar(self.dl, parent=getattr(self, 'mbar', None), leave=False)
        self.pbar.update(0)
        
    def after_batch(self):
        self.pbar.update(self.iter_num+1)
    
    def after_train(self):
        self.pbar.on_iter_end()
        
    def after_validate(self):
        self.pbar.on_iter_end()
        
    def after_fit(self):
        if getattr(self, 'mbar', False):
            self.mbar.on_iter_end()
            delattr(self, 'mbar')
        if hasattr(self, 'old_logger'): self.learn.logger = self.old_logger

    def _write_stats(self, log):
        if getattr(self, 'mbar', False): self.mbar.write([f'{l:.6f}' if isinstance(l, float) else str(l) for l in log], table=True)
    

    

# %% ../../nbs/11_l2r.callbacks.ipynb 34
class SaveCallBack(Monitor):
    order = Monitor.order+1
    def __init__(self, 
        fname, 
        monitor='ndcg_at_6', 
        comp=None, 
        min_delta=0., 
        reset_on_fit=False,
    ):
        super().__init__(monitor=monitor, comp=comp, min_delta=min_delta, reset_on_fit=reset_on_fit)
        self.last_saved_path = None
        store_attr('fname')
        
    @property
    def best(self): return self._best
    @best.setter    
    def best(self, b): self._best = b
    
    def after_epoch(self):
        super().after_epoch()
        if self.new_best:
            print(f'Better model found at epoch {self.epoch} with {self.monitor} value: {self.best}.')
            self.learn.save(self.fname)
    
    # def after_fit(self):
        # if self.best_at_end: self.learn.load(self.fname)

# %% ../../nbs/11_l2r.callbacks.ipynb 36
class InterruptCallback(Callback):
    def __init__(self, epoch):
        self._interupt_before = epoch
    def before_epoch(self):
        if self.epoch == self._interupt_before:
            raise CancelFitException

# %% ../../nbs/11_l2r.callbacks.ipynb 38
class CSVLog(Callback):
    "Log the results displayed in `learn.path/fname`"
    order=60
    def __init__(self, fname='history.csv', append=False):
        self.fname,self.append = Path(fname),append

    def read_log(self):
        "Convenience method to quickly access the log."
        return pd.read_csv(self.path/self.fname)

    def before_fit(self):
        "Prepare file with metric names."
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.file = (self.path/self.fname).open('a' if self.append else 'w')
        self.file.write(','.join(self.track_results.metric_names) + '\n')
        self.old_logger,self.learn.logger = self.logger,self._write_line

    def _write_line(self, log):
        "Write a line with `log` and call the old logger."
        self.file.write(','.join([str(t) for t in log]) + '\n')
        self.file.flush()
        os.fsync(self.file.fileno())
        self.old_logger(log)

    def after_fit(self):
        "Close the file and clean up."
        self.file.close()
        self.learn.logger = self.old_logger

# %% ../../nbs/11_l2r.callbacks.ipynb 40
from torch.cuda.amp import GradScaler, autocast

# %% ../../nbs/11_l2r.callbacks.ipynb 41
class AMPMode(Enum):
    "Automatic mixed precision modes for ease of completion"
    FP16 = 'fp16'
    BF16 = 'bf16'

# %% ../../nbs/11_l2r.callbacks.ipynb 42
class TestCallback(Callback):
    order = 1000
    def before_batch(self):
        # import pdb; pdb.set_trace()
        ic(self.opt.hypers)
    # def after_epoch(self):
        # ic(self.opt.hypers)

# %% ../../nbs/11_l2r.callbacks.ipynb 43
class MemoryErrorTest(Callback):
    order=SaveCallBack.order-1
    def after_batch(self):
        if self.iter_num == 4:
            # raise torch.cuda.OutOfMemoryError("Memory !!!")
            raise CancelTrainException()
    def after_cancel_train(self): pass

# %% ../../nbs/11_l2r.callbacks.ipynb 44
@delegates(GradScaler)
class MixedPrecision(Callback):
    "Mixed precision training using Pytorch's Automatic Mixed Precision (AMP)"
    order = 10
    def __init__(self,
        amp_mode:str|AMPMode=AMPMode.FP16, # Mixed Precision training mode. Supports fp16 and bf16.
        **kwargs
    ):
        amp_mode = AMPMode(amp_mode)
        store_attr(names='amp_mode')
        self.kwargs = kwargs

    def before_fit(self):
        if self.amp_mode == AMPMode.BF16:
            if torch.cuda.is_available() and not torch.cuda.is_bf16_supported():
                raise ValueError("Unsupported GPU for bfloat16 mixed precision training")
            dtype = torch.bfloat16
        elif self.amp_mode == AMPMode.FP16:
            dtype = torch.float16
        else:
            raise ValueError(f"Unrecognized precision: {self.amp_mode}")
        # `GradScaler` is not needed for bfloat16 as fp32 and bf16 have the same range
        self.kwargs['enabled'] = dtype == torch.float16
        self.autocast,self.learn.scaler,self.scales = autocast(dtype=dtype), GradScaler(**self.kwargs),L()

    def before_batch(self): 
        if self.model.training: 
            self.autocast.__enter__()
    def after_pred(self):
        pass
        # assert self.learn.preds.dtype == torch.float16
        # self.learn.preds = to_float(self.preds)
    def after_loss(self): self.autocast.__exit__(None, None, None)
    def before_backward(self): 
        # pass
        # import pdb; pdb.set_trace()
        # self.learn.preds = self.scaler.scale(self.preds)
        assert self.learn.preds.dtype == torch.float16
        # assert self.learn.lambda_i.dtype == torch.float16
        assert self.learn.srtd_preds.dtype == torch.float16
        self.learn.srtd_preds = self.scaler.scale(self.srtd_preds)
        # self.learn.lambda_i = self.scaler.scale(self.lambda_i)
    # def after_backward(self):
    #     print(self.track_results.grads_full['token_factors.weight']
    #     import pdb; pdb.set_trace()
    def before_step(self):
        "Use `self` as a fake optimizer. `self.skipped` will be set to True `after_step` if gradients overflow."
        self.skipped=True
        self.scaler.step(self)
        if self.skipped: raise CancelStepException()
        self.scales.append(self.scaler.get_scale())
        # pass
    def after_step(self): self.learn.scaler.update()
    def after_cancel_step(self): pass
    def after_fit(self): self.autocast,self.learn.scaler,self.scales = None,None,None

    @property
    def param_groups(self):
        "Pretend to be an optimizer for `GradScaler`"
        return self.opt.param_groups
    def step(self, *args, **kwargs):
        "Fake optimizer step to detect whether this batch was skipped from `GradScaler`"
        self.skipped=False
    def after_epoch(self): 
        torch.cuda.empty_cache()
    def after_batch(self): 
        torch.cuda.empty_cache()

# %% ../../nbs/11_l2r.callbacks.ipynb 47
class GradientClip(Callback):
    "Clip norm of gradients"
    order=MixedPrecision.order+1
    def __init__(self,max_norm:float=1., norm_type:float=2.0): store_attr()
    def before_step(self): 
        # import pdb; pdb.set_trace()
        nn.utils.clip_grad_norm_(self.model.parameters(), self.max_norm, self.norm_type)
   



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/15_l2r.data.stat.ipynb.

# %% auto 0
__all__ = []

# %% ../../../nbs/15_l2r.data.stat.ipynb 2
# from fastcore.basics import *
# from fastai.torch_core import *
# from fastai.data.core import *
# from fastai.data.transforms import *
# from fastai.text.core import *
# from fastai.text.data import *
# from xcube.imports import *
# from xcube.torch_imports import *



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/06_data.transforms.ipynb.

# %% auto 0
__all__ = ['ListToTensor']

# %% ../../nbs/06_data.transforms.ipynb 2
from fastai.torch_basics import *
from fastai.data.core import *
from fastai.data.load import *
from fastai.data.all import *

# %% ../../nbs/06_data.transforms.ipynb 5
class ListToTensor(DisplayedTransform):
    def encodes(self, x): return tensor(x).long()
    def decodes(self, x): return L(x, use_list=True)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/08_l2r.gradients.ipynb.

# %% auto 0
__all__ = ['rank_loss2', 'rank_loss3', 'loss_fn', 'loss_fn2']

# %% ../../nbs/08_l2r.gradients.ipynb 2
from fastai.torch_imports import *
from ..imports import *

# %% ../../nbs/08_l2r.gradients.ipynb 8
def _summation(sl, ij):
    sumer = []
    for i in range(sl):
        _x = torch.nonzero(ij == i, as_tuple=False)
        _x[:, -1] = torch.pow(-1, _x[:, 1])
        sumer.append(_x)
    return torch.stack(sumer, dim=0)

def _idcg(xb, k=None, gain_fn=None):
    # pdb.set_trace()
    x = xb[:, :, :, -1]
    ranks = x.argsort(dim=-1, descending=True).argsort(dim=-1) # ranking by the scores, highest score gets rank 0
    dfs = 1/torch.log2(ranks + 2)
    gains = torch.pow(2, x) if gain_fn == 'exp' else torch.pow(x, 3)
    idg = gains * dfs
    idcg = idg.sum(dim=-1)
    
    idcg_at_k = None
    if k is not None:
        topk, topk_idxs = torch.topk(x, k=k, dim=-1, largest=True)
        # topk_relvs = torch.take_along_dim(x, topk_idxs, dim=-1)
        dfs_at_k = 1/torch.log2(2 + torch.arange(k)).cuda()
        gains_at_k = torch.pow(2, topk) if gain_fn == 'exp' else torch.pow(topk, 3)
        idg_at_k = gains_at_k * dfs_at_k
        idcg_at_k = idg_at_k.sum(-1)
    
    return idcg, idcg_at_k

# %% ../../nbs/08_l2r.gradients.ipynb 9
def rank_loss2(preds, xb, sigma=0.5, lambrank=False, gain_fn=None, k=6):
    # In the following `ij` is essentially the set $I$
    sl = xb.shape[2]
    ij = torch.as_tensor(np.fromiter(itertools.combinations(np.arange(sl), 2), dtype=np.dtype((int,2))),
                                device=xb.device)#.expand(xb.shape[0], xb.shape[1], -1, -1)
    
    # Sort the tokens by the model prediction scores so that we can compute the set $I$ defined above:
    srtd_preds, srtd_idxs = preds[:, :, :,  0].sort(descending=True)
    
    srtd_ranks = srtd_preds.new_empty(srtd_preds.size())#srtd_idxs.argsort()
    srtd_ranks[:,:] = torch.arange(preds.shape[2])
    ri_rj = srtd_ranks[:, :, ij] # these are the ranks for token_i and token_j
    dfi_dfj = 1.0 / torch.log2(ri_rj + 2)
    dfi = dfi_dfj[:,:,:,0]
    dfj = dfi_dfj[:,:,:,1]
        
    srtd_relvs = torch.take_along_dim(xb[:, :, :, -1], srtd_idxs, dim=-1)
    pi_pj = srtd_preds[:, :, ij] # these are p_i and p_j 
    pi, pj = pi_pj[:, :, :, 0], pi_pj[:, :, :, 1]
    exp_ij = torch.exp(sigma * (pi - pj))
    si_sj = srtd_relvs[:, :, ij] # these are the relevance scores for token_i and token_j
    si, sj= si_sj[:, :, :, 0], si_sj[:, :, :, 1]
    gain_i, gain_j = ( torch.pow(2.0, si), torch.pow(2.0, sj) ) if gain_fn == 'exp' else ( torch.pow(si, 3.0), torch.pow(sj, 3.0) ) # cubic
    signs = torch.sign(si - sj)
    delta_dcg = torch.abs((gain_i - gain_j) * (dfi - dfj))
    idcg, idcg_at_k = _idcg(xb, k=k, gain_fn=gain_fn)
    delta_ndcg_at_k = delta_dcg / idcg_at_k.unsqueeze(-1)
    
    lambda_ij = sigma * (  0.5 * (1 - signs) -  1/(1 + exp_ij) )
    if lambrank: lambda_ij *= delta_ndcg_at_k # use this for Lambda-Rank
    
    sumer = _summation(sl, ij)
    idxr, signs = sumer[:, :, 0], sumer[:, :, -1]
    # Now we can compute $\lambda_i$ from eq: 4,
    lambda_i = (lambda_ij[:, :, idxr] * signs).sum(dim=-1)
    
    return srtd_preds, lambda_i

# %% ../../nbs/08_l2r.gradients.ipynb 10
def rank_loss3(preds, xb, sigma=0.5, lambrank=False, gain_fn=None, k=6):
    with torch.no_grad():
        x = xb[:, :, :, -1, None]
        x_t = xb[:, :, :, -1, None].transpose(-1,-2)
        preds_t = preds.transpose(-1,-2)
        preds_rank = preds[:, :, :, 0].argsort(dim=-1, descending=True).argsort(dim=-1).unsqueeze(-1)
        preds_rank_t = preds_rank.transpose(-1,-2)
        
        exp_ij= 1.0 + torch.exp(sigma* (preds - preds_t)).half()
        rel_diff = x - x_t
        gain_diff = torch.pow(2.0, x) - torch.pow(2.0, x_t) if gain_fn == 'exp' else torch.pow(x, 3.0) - torch.pow(x_t, 3.0)
        decay_diff = 1.0/torch.log2(preds_rank + 2.0) - 1.0/torch.log2(preds_rank_t  + 2.0)
        idcg, idcg_at_k = _idcg(xb, k=k, gain_fn=gain_fn)
        idcg_at_k = idcg_at_k[..., None, None]
        # pdb.set_trace()
        delta_ndcg_at_k = torch.abs(gain_diff * decay_diff * 1/idcg_at_k).half()
        pos_pairs = (rel_diff > 0).to(torch.int8) #pos_pairs = (rel_diff > 0).float()
        neg_pairs = (rel_diff < 0).to(torch.int8) #neg_pairs = (rel_diff < 0).float()
        # pdb.set_trace()
        S_ij = pos_pairs - neg_pairs # S_ij = (pos_pairs - neg_pairs).half()
        lambda_update = sigma * (  0.5 * (1 - S_ij) -  (1/exp_ij.half()) )
        if lambrank: lambda_update *= delta_ndcg_at_k 
        lambda_update = lambda_update.sum(dim=-1, keepdim=True)
        # free memory
        # del preds_t, preds_rank, preds_rank_t, exp_ij, rel_diff, gain_diff, decay_diff, idcg, idcg_at_k, delta_ndcg_at_k, pos_pairs, neg_pairs, S_ij
        exp_ij, rel_diff, gain_diff, decay_diff, idcg, idcg_at_k, delta_ndcg_at_k, pos_pairs, neg_pairs, S_ij = None, None, None, None, None, None, None, None, None, None
        # S_ij = None
        import gc; gc.collect(); torch.cuda.empty_cache()
    return preds, lambda_update

# %% ../../nbs/08_l2r.gradients.ipynb 12
def loss_fn(preds, xb, sigma=0.5):
    
    srtd_relvs, srtd_idxs = xb[:, :, :, -1].sort(descending=True)
    srtd_preds = torch.take_along_dim(preds[:,:,:,0], srtd_idxs, dim=-1)

    sl = torch.arange(xb.shape[2], device=xb.device)
    ij = torch.cartesian_prod(sl, sl)
    idxs, = torch.nonzero(ij[:, 0] < ij[:, 1], as_tuple=True)
    ij = ij[idxs]
    
    si_sj = srtd_relvs[:, :, ij] # these are the relevance scores for token_i and token_j
    si, sj= si_sj[:, :, :, 0], si_sj[:, :, :, 1]
    signs = torch.sign(si - sj)
    pi_pj = srtd_preds[:, :, ij]
    pi, pj = pi_pj[:,:,:,0], pi_pj[:,:,:,1]
    exp_ij = torch.exp(-sigma*(pi -pj))
    exp_ij[exp_ij==torch.inf] = tensor(1e6)
    C = ( 0.5*(1 - signs)*sigma*(pi -pj) + torch.log(1 + exp_ij) ) #shape (64, 2234, 64)
    # C = C.sum(dim=-1) # shape (64, 2234)
    C = C.mean(dim=-1)
    return C#.mean()

def loss_fn2(preds, xb, sigma=.5):
    "Computes average pairwise cross-entropy loss"
    sl = xb.shape[2]
    rel_diff = xb[:, :, :, -1, None] - xb[:, :, :, -1, None].transpose(-1, -2)
    pos_pairs = (rel_diff > 0).float()
    neg_pairs = (rel_diff < 0).float()
    S_ij = pos_pairs - neg_pairs
    preds_diff = preds - preds.transpose(-1, -2)
    C = .5 * (1 - S_ij) * sigma * preds_diff - F.logsigmoid(sigma * preds_diff)
    C = torch.triu(C, diagonal=1) # to take each pair only once
    C = C.sum((-1,-2)) / (C.new_ones(C.shape[-2:]).triu(diagonal=1).sum())
    return C



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/19_text.callbacks.ipynb.

# %% auto 0
__all__ = ['LabelForcing', 'RunvalCallback']

# %% ../../nbs/19_text.callbacks.ipynb 1
from fastai.torch_imports import *
from fastai.torch_core import *
from fastai.callback.core import *
from fastcore.all import *
from ..imports import *
from ..metrics import *

# %% ../../nbs/19_text.callbacks.ipynb 6
class LabelForcing(Callback):
    def __init__(self, end_epoch):
        self.end_epoch = end_epoch
    
    @torch.no_grad()
    def after_pred(self):
        if self.training and self.epoch <= self.end_epoch:
            pred, attn, wgts = self.learn.pred
            pos = Tensor(self.y) == 1
            # pred[pos] += 3*pred.std()
            # pred[~pos] -= 3*pred.std()
            attn[pos] += attn[pos].std(dim=-1).unsqueeze(-1)
            attn[~pos] -= attn[~pos].std(dim=-1).unsqueeze(-1)

# %% ../../nbs/19_text.callbacks.ipynb 8
from fastai.callback.progress import *
from fastai.learner import Recorder

# %% ../../nbs/19_text.callbacks.ipynb 9
class _FakeLearner: 
    def to_detach(self,b,cpu=True,gather=True):
        return to_detach(b,cpu,gather)

# %% ../../nbs/19_text.callbacks.ipynb 10
def _cpupy(t): return t.cpu().numpy() if isinstance(t, Tensor) else t

# %% ../../nbs/19_text.callbacks.ipynb 11
class RunvalCallback(Callback):
    order=ProgressCallback.order-1
    
    def __init__(self, mets):
        self.mets = mets
        
    def before_train(self): 
        self.val_cyclit = itertools.cycle(self.dls.valid) 
        self._fake_l = _FakeLearner()
        self.mets.map(Self.reset())
        self.len = len(self.dls.valid)
        self._launch_tqbar()
        
    def _launch_tqbar(self):
        if hasattr(self, 'tqbar'): self.tqbar.close()
        self.mets.map(Self.reset())
        self.counter = 0
        self.tqbar = tqdm(total=self.len, leave=False)
        self.tqbar.set_description('Running validation')
        
    def after_batch(self):
        if not self.training: return
        self.model.eval()
        self.learn.training=False
        self._batch_val()
        self.model.train()
        self.learn.training=True
        
    @torch.no_grad()
    def _batch_val(self):
        xb, yb = next(self.val_cyclit)
        self.tqbar.update(1)
        if self.counter == self.len:
            self._launch_tqbar()
        else: self.counter += 1
        self._fake_l.yb = (yb,)
        self._fake_l.pred, *_ = self.model(xb) 
        self._fake_l.loss = Tensor(self.loss_func(self._fake_l.pred, yb))
        for met in self.mets: met.accumulate(self._fake_l)
        # pdb.set_trace()
        comment = dict(self.mets.attrgot('name').zipwith(self.mets.attrgot('value').map(_cpupy)))
        self.tqbar.set_postfix(comment)
    
    def after_train(self):
        if hasattr(self, 'val_cyclit'): delattr(self, 'val_cyclit')
        if hasattr(self, 'tqbar'): self.tqbar.close(); delattr(self, 'tqbar')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_collab.ipynb.

# %% auto 0
__all__ = ['match_embeds', 'load_pretrained_keys', 'CollabLearner', 'collab_learner']

# %% ../nbs/05_collab.ipynb 2
from fastai.tabular.all import *
from fastai.collab import *

# %% ../nbs/05_collab.ipynb 8
def match_embeds(
    old_wgts:dict, # Embedding weights of the pretrained model
    old_vocab:list, # Vocabulary (tokens and labels) of the corpus used for pretraining
    new_vocab:dict # Current collab corpus vocabulary (`users` and `items`)
) -> dict:
    """
    Convert the `users` and `items` (possibly saved as `0.module.encoder.weight` and `1.attn.lbs_weight.weight` respectively) 
    embedding in `old_wgts` to go from `old_vocab` to `new_vocab`
    """
    u_bias, u_wgts = None, old_wgts.get('0.module.encoder.weight')
    i_bias, i_wgts = old_wgts.get('1.attn.lbs_bias.weight', None), old_wgts.get('1.attn.lbs_weight.weight')
    u_wgts_m, i_wgts_m = u_wgts.mean(0), i_wgts.mean(0)
    new_u_wgts = u_wgts.new_zeros((len(new_vocab['token']), u_wgts.size(1)))
    new_i_wgts = i_wgts.new_zeros((len(new_vocab['label']), i_wgts.size(1)))
    if u_bias is not None:
        u_bias_m = u_bias.mean(0)
        new_u_bias = u_bias.new_zeros((len(new_vocab['token']), 1))
    if i_bias is not None:
        i_bias_m = i_bias.mean(0)
        new_i_bias = i_bias.new_zeros((len(new_vocab['label']), 1))
    u_old = old_vocab[0]
    u_old_o2i = u_old.o2i if hasattr(u_old, 'o2i') else {w:i for i,w in enumerate(u_old)}
    i_old = old_vocab[1]
    i_old_o2i = i_old.o2i if hasattr(i_old, 'o2i') else {w:i for i,w in enumerate(i_old)}
    u_miss, i_miss = 0, 0
    for i,w in enumerate(new_vocab['token']):
        idx = u_old_o2i.get(w, -1)
        new_u_wgts[i] = u_wgts[idx] if idx>=0 else u_wgts_m
        if u_bias is not None: new_u_bias[i] = u_bias[idx] if idx>=0 else u_bias_m
        if idx == -1: u_miss = u_miss + 1
    for i,w in enumerate(new_vocab['label']):
        idx = i_old_o2i.get(w, -1)
        new_i_wgts[i] = i_wgts[idx] if idx>=0 else i_wgts_m
        if i_bias is not None: new_i_bias[i] = i_bias[idx] if idx>=0 else i_bias_m
        if idx == -1: i_miss = i_miss + 1
    old_wgts['0.module.encoder.weight'] = new_u_wgts
    if '0.module.encoder_dp.emb.weight' in old_wgts: old_wgts['0.module.encoder_dp.emb.weight'] = new_u_wgts.clone()
    if u_bias is not None: pass
    old_wgts['1.attn.lbs_weight.weight'] = new_i_wgts
    if '1.attn.lbs_weight_dp.emb.weight' in old_wgts: old_wgts['1.attn.lbs_weight_dp.emb.weight'] = new_i_wgts.clone()
    if i_bias is not None: old_wgts['1.attn.lbs_bias.weight'] = new_i_bias
    return old_wgts, u_miss, i_miss

# %% ../nbs/05_collab.ipynb 10
def load_pretrained_keys(
    model, # Model architecture
    wgts:dict # Model weights
) -> tuple:
    "Load relevant pretrained `wgts` in `model"
    sd = model.state_dict()
    u_wgts, u_bias = wgts.get('0.module.encoder.weight', None), None
    if u_wgts is not None: sd['u_weight.weight'].data = u_wgts.data
    if u_bias is not None: sd['u_bias.weight'].data = u_bias.data
    i_wgts, i_bias = wgts.get('1.attn.lbs_weight.weight', None), wgts.get('1.attn.lbs_bias.weight', None)
    if i_wgts is not None: sd['i_weight.weight'].data = i_wgts.data
    if i_bias is not None: sd['i_bias.weight'].data = i_bias.data
    return model.load_state_dict(sd)

# %% ../nbs/05_collab.ipynb 11
class CollabLearner(Learner):
    "Basic class for a `Learner` in Collab."
    @delegates(save_model)
    def save(self, 
        file:str, # Filename for the state_directory of model
        **kwargs):
        """
        Save model and optimizer state (if `with_opt`) to `self.path/self.model_dir/file`
        Save `self.dls.classes` to `self.path.self.model_dir/collab_vocab.pkl`
        """
        model_file = join_path_file(file, self.path/self.model_dir, ext='.pth')
        vocab_file = join_path_file(file+'_vocab', self.path/self.model_dir, ext='.pkl')
        save_model(model_file, self.model, getattr(self,'opt', None), **kwargs)
        save_pickle(vocab_file, self.dls.classes)
        return model_file
    
    def load_vocab(self,
        wgts_fname:str, #Filename of the saved weights
        vocab_fname:str, # Saved vocabulary filename in pickle format
        model=None # Model to load parameters from, deafults to `Learner.model`
    ):
        "Load the vocabulary (`users` and/or `items`) from a pretrained model and adapt it to the collab vocabulary."
        old_vocab = load_pickle(vocab_fname)
        new_vocab = self.dls.classes
        distrib_barrier()
        wgts = torch.load(wgts_fname, map_location=lambda storage,loc: storage)
        if 'model' in wgts: wgts = wgts['model'] # Just in case the pretrained model was saved with an optimizer
        wgts, *_ = match_embeds(wgts, old_vocab, new_vocab)
        load_pretrained_keys(self.model if model is None else model, wgts)
        return self

# %% ../nbs/05_collab.ipynb 15
@delegates(Learner.__init__)
def collab_learner(dls, n_factors=50, use_nn=False, emb_szs=None, layers=None, config=None, y_range=None, loss_func=None, pretrained=False, **kwargs):
    "Create a Learner for collaborative filtering on `dls`."
    emb_szs = get_emb_sz(dls, ifnone(emb_szs, {}))
    if loss_func is None: loss_func = MSELossFlat()
    if config is None: config = tabular_config()
    if y_range is not None: config['y_range'] = y_range
    if layers is None: layers = [n_factors]
    if use_nn: model = EmbeddingNN(emb_szs=emb_szs, layers=layers, **config)
    else:      model = EmbeddingDotBias.from_classes(n_factors, dls.classes, y_range=y_range)
    learn = CollabLearner(dls, model, loss_func=loss_func, **kwargs)
    if pretrained:
        try: fnames = [list(learn.path.glob(f'**/clas/*clas*.{ext}'))[0] for ext in ['pth', 'pkl']] 
        except: IndexError: print(f'The model in {learn.path} is incomplete, re-train it'); raise
        learn = learn.load_vocab(*fnames, model=learn.model)
    return learn



from ..torch_imports import *
from ..imports import *
from ..utils import *
from ..metrics import *
from ..data import *
from ..data.all import *
from .learner import *
from .callbacks import *
from .models.all import *


# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/09_l2r.learner.ipynb.

# %% auto 0
__all__ = ['SkipToEpoch', 'L2RLearner', 'get_learner']

# %% ../../nbs/09_l2r.learner.ipynb 2
from fastai.torch_imports import *
from fastai.learner import *
from fastai.optimizer import *
from fastai.torch_core import *
from fastcore.all import *
from ..imports import *
from ..metrics import *
from .gradients import *

# %% ../../nbs/09_l2r.learner.ipynb 6
from fastai.callback.core import *

# %% ../../nbs/09_l2r.learner.ipynb 7
class SkipToEpoch(Callback):
    "Skip training up to `epoch`"
    order = 70
    
    def __init__(self, epoch:int):
        self._skip_to = epoch

    def before_epoch(self):
        if self.epoch < self._skip_to:
            raise CancelEpochException
    
    def after_cancel_epoch(self): pass

# %% ../../nbs/09_l2r.learner.ipynb 8
class L2RLearner:
    def __init__(self, 
        model, 
        dls, 
        grad_func, 
        loss_func, 
        lr, 
        cbs, 
        opt_func=SGD, 
        path=None,
        model_dir:str|Path='models', # Subdirectory to save and load models
        moms:tuple=(0.95,0.08,0.95)
    ):
        store_attr(but='cbs')
        self.path = Path(path) if path is not None else getattr(dls, 'path', Path('.'))
        self.cbs = L()
        self.add_cbs(cbs)
        self.logger = print

    def add_cb(self, cb):
        cb.learn = self
        setattr(self, cb.name, cb)
        self.cbs.append(cb)
        return self

    def add_cbs(self, cbs):
        L(cbs).map(self.add_cb)
        return self
    
    @contextmanager
    def added_cbs(self, cbs):
        self.add_cbs(cbs)
        try: yield
        finally: self.remove_cbs(cbs)
        
    @contextmanager
    def removed_cbs(self, cbs):
        self.remove_cbs(cbs)
        try: yield self
        finally: self.add_cbs(cbs)
        
    def remove_cbs(self, cbs):
        L(cbs).map(self.remove_cb)
        return self
    
    def remove_cb(self, cb):
        cb.learn = None
        if hasattr(self, cb.name): delattr(self, cb.name)
        if cb in self.cbs: self.cbs.remove(cb)
        return self

    def _step(self): self.opt.step()
        
    def one_batch(self, *args, **kwargs):
        # self('before_batch')
        self.preds = self.model(self.xb)
        self('after_pred')
        # if self.model.training: # training
        if not self.model.training: return
        self.srtd_preds, self.lambda_i = self.grad_func(self.preds, self.xb)
        self.lambda_i = self.lambda_i.half()
        self('after_loss')
        self('before_backward')
        self.srtd_preds.backward(self.lambda_i)
        self('after_backward')
        
        # free memory (TODO: Put this in a little callback)
        self.lambda_i = None
        import gc; gc.collect()
        torch.cuda.empty_cache()
            
        # self('before_step')
        # self.opt.step()
        self._with_events(self._step, 'step', CancelStepException)
        # self('after_step')
        self.opt.zero_grad()
            
        # self('after_batch')
        
    def one_epoch(self, train, **kwargs):
        self.model.training = train
        self.dl = self.dls.train if train else self.dls.valid
        (self._do_epoch_validate, self._do_epoch_train)[self.model.training](**kwargs)
        
    def _do_epoch_train(self, *args, **kwargs):
        self._with_events(partial(self._all_batches, *args, **kwargs), 'train', CancelTrainException)
        # self('before_train')
        # self._all_batches(*args, **kwargs)
        # self('after_train')
        
    def _do_epoch_validate(self, *args, idx=1, dl=None, **kwargs):
        if dl is None: dl = self.dls[idx]
        self.dl = dl
        with torch.no_grad():
            self._with_events(partial(self._all_batches, *args, **kwargs), 'validate', CancelValidException)
            # self('before_validate')
            # self._all_batches(*args, **kwargs)
            # self('after_validate')
        
    def _all_batches(self, *args, **kwargs):
        self.n_iter = len(self.dl)
        for self.iter_num, self.xb in enumerate(self.dl):
            self._with_events(partial(self.one_batch, *args, **kwargs), 'batch', CancelBatchException)
    
    def create_opt(self):
        self.opt = self.opt_func(self.model.parameters(), self.lr)
        # self.opt.clear_state()
        return self.opt
    
    def _do_epoch(self, **kwargs):
        self.one_epoch(True, **kwargs)
        self.one_epoch(False, **kwargs)

    def fit(self, n_epochs, cbs=None, start_epoch=0, lr=None, wd=None, reset_opt=False, **kwargs):
        if start_epoch != 0:
            cbs = L(cbs) + SkipToEpoch(start_epoch)
        with self.added_cbs(cbs):
            opt = getattr(self, 'opt', None)
            if opt is None or reset_opt: self.create_opt()
            if wd is not None: self.opt.set_hypers(wd=wd)
            self.opt.set_hypers(lr=self.lr if lr is None else lr)
            self.n_epochs = n_epochs
            self('before_fit')
            try:
                for self.epoch,_ in enumerate(range(self.n_epochs)):
                    self._with_events(partial(self._do_epoch, **kwargs), 'epoch', CancelEpochException)
                    # self('before_epoch')
                    # self.one_epoch(True, **kwargs)
                    # self.one_epoch(False, **kwargs)
                    # self('after_epoch')
            except CancelFitException: pass 
            self('after_fit')
    
    def validate(self, idx=1, dl=None, **kwargs):
        try: 
            self.model.training = False
            self._do_epoch_validate(idx, dl, **kwargs)
        except CancelFitException: pass
    
    def __call__(self, name):
        for cb in self.cbs: getattr(cb, name, noop)()
        
    def _with_events(self, f, event_type, ex, final=noop):
        try: self(f'before_{event_type}'); f()
        except ex: self(f'after_cancel_{event_type}')
        self(f'after_{event_type}'); final()
        

# %% ../../nbs/09_l2r.learner.ipynb 10
@patch
@delegates(save_model)
def save(self:L2RLearner, file, **kwargs):
    "Save model and optimizer state (if 'with_opt') to `self.path/file`"
    file = join_path_file(file, self.path/self.model_dir, ext='.pth')
    save_model(file, self.model, getattr(self, 'opt', None), **kwargs)
    return file

# %% ../../nbs/09_l2r.learner.ipynb 11
@patch
@delegates(load_model)
def load(self:L2RLearner, file, device=None, **kwargs):
    "Load model and optimizer state (if `with_opt`) from `self.path/file` using `device`"
    if device is None and hasattr(self.dls, 'device'): device = self.dls.device
    self.opt = getattr(self, 'opt', None)
    if self.opt is None: self.create_opt()
    file = join_path_file(file, self.path/self.model_dir, ext='.pth')
    load_model(file, self.model, self.opt, device=device, **kwargs)
    return self

# %% ../../nbs/09_l2r.learner.ipynb 12
@patch
def show_results(self:L2RLearner, device=None, k=None):
    "Produces the ranking for 100 random labels"
    dataset = to_device(self.dls.train.dataset, device=device)
    num_lbs = dataset.shape[0]
    idxs = torch.randperm(num_lbs)[:100]
    xb = dataset[idxs]
    xb = xb.unsqueeze(0)
    preds, preds_rank, *_,  _ndcg_at_k = ndcg(self.model(xb), xb, k=k)
    if _ndcg_at_k is not None: _ndcg_at_k.squeeze_(0) 
    # lbs = xb[:, :, :, 1].unique().cpu().numpy()
    lbs = idxs.numpy()
    cols = pd.MultiIndex.from_product([lbs, ('tok', 'lbl', 'rank', 'score', 'preds', 'model_rank')], names=['label', 'key2'])
    data = torch.concat( (xb, preds.unsqueeze(-1), preds_rank.unsqueeze(-1)), dim=-1).squeeze(0).permute(1, 0, 2).contiguous()
    data = data.reshape(data.shape[0], -1)
    df_results = pd.DataFrame(data, columns=cols)
    df_results.index.name = 'toks'
    # pd.set_option('display.max_columns', None)
    df_ndcg = pd.DataFrame({'labels': lbs, 'ndcg_at_k':_ndcg_at_k.cpu().numpy()})
    return df_results, df_ndcg

# %% ../../nbs/09_l2r.learner.ipynb 14
def get_learner(model, dls, grad_fn=rank_loss3, loss_fn=loss_fn2, lr=1e-5, cbs=None, opt_func=partial(SGD, mom=0.9), lambrank=False, **kwargs):
    if lambrank: grad_fn = partial(grad_fn, lambrank=lambrank)
    learner = L2RLearner(model, dls, grad_fn, loss_fn, lr, cbs, opt_func=opt_func, **kwargs)
    return learner



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_layers.ipynb.

# %% ../nbs/01_layers.ipynb 2
from __future__ import annotations
from typing import Union
from fastai.imports import *
from fastai.torch_imports import *
from fastai.torch_core import *
from fastai.layers import *
from fastai.text.models.awdlstm import EmbeddingDropout, RNNDropout

from .utils import *

# %% auto 0
__all__ = ['ElemWiseLin', 'LinBnFlatDrop', 'LinBnDrop', 'Embedding', 'Linear_Attention', 'Planted_Attention', 'PlantedLMDecoder',
           'Diffntble_Planted_Attention', 'lincomb', 'split_sort', 'XMLAttention']

# %% ../nbs/01_layers.ipynb 13
def _create_bias(size, with_zeros=False):
    if with_zeros: return nn.Parameter(torch.zeros(*size))
    return nn.Parameter(torch.zeros(*size).uniform_(-0.1, 0.1))

# %% ../nbs/01_layers.ipynb 14
class ElemWiseLin(Module):
    initrange=0.1
    def __init__(self, dim0, dim1, add_bias=False, **kwargs):
        store_attr()
        self.lin = nn.Linear(dim1, dim0, **kwargs)
        # init_default(self.lin, func=partial(torch.nn.init.uniform_, a=-self.initrange, b=self.initrange))
        init_default(self.lin)
        if self.add_bias: self.bias = _create_bias((1, ))
        
    def forward(self, x):
        res = torch.addcmul(self.bias if self.add_bias else x.new_zeros(1), x, self.lin.weight)# * self.lin.weight
        return res #+ self.bias if self.add_bias else res

# %% ../nbs/01_layers.ipynb 17
class LinBnFlatDrop(nn.Sequential):
    "Module grouping `BatchNorm1dFlat`, `Dropout` and `Linear` layers"
    def __init__(self, n_in, n_out, bn=True, p=0., act=None, lin_first=False):
        layers = [BatchNorm1dFlat(n_out if lin_first else n_in)] if bn else []
        if p != 0: layers.append(nn.Dropout(p))
        lin = [nn.Linear(n_in, n_out, bias=not bn)]
        if act is not None: lin.append(act)
        layers = lin+layers if lin_first else layers+lin
        super().__init__(*layers)

# %% ../nbs/01_layers.ipynb 18
class LinBnDrop(nn.Sequential):
    "Module grouping `BatchNorm1d`, `Dropout` and `Linear` layers"
    def __init__(self, n_in, n_out=None, bn=True, ln=True, p=0., act=None, lin_first=False, ndim=1):
        if not ln and lin_first: raise Exception(AssertionError)
        layers = [BatchNorm(n_out if ln and lin_first else n_in, ndim=ndim)] if bn else []
        if p != 0: layers.append(nn.Dropout(p))
        lin = [nn.Linear(n_in, n_out, bias=not bn)] if ln else []
        if ln and act is not None: lin.append(act)
        layers = lin+layers if lin_first else layers+lin
        super().__init__(*layers)

# %% ../nbs/01_layers.ipynb 24
class Embedding(nn.Embedding):
    "Embedding layer with truncated normal initialization"
    def __init__(self, ni, nf, std=0.01, **kwargs):
        super().__init__(ni, nf, **kwargs)
        trunc_normal_(self.weight.data, std=std)

# %% ../nbs/01_layers.ipynb 26
def _linear_attention(sentc:Tensor, # Sentence typically `(bs, bptt, nh)` output of `SentenceEncoder`
                      based_on: nn.Embedding|Module # xcube's `Embedding(n_lbs, nh)` layer holding the label embeddings or a full fledged model
                  ):
    return sentc @ based_on.weight.transpose(0,1)

# %% ../nbs/01_layers.ipynb 28
def _planted_attention(sentc: Tensor, # Sentence typically `(bs, bptt)` containing the vocab idxs that goes inside the encoder
                       brain: Tensor # label specific attn wgts for each token in vocab, typically of shape `(vocab_sz, n_lbs)`
                     ):
    return brain[sentc.long()]

# %% ../nbs/01_layers.ipynb 30
def _diffntble_planted_attention(sentc_dec: Tensor, # Sentence `(bs, bptt)` typically containing the vocab idxs obtained after decoding what comes out of the encoder
                         l2r: nn.ModuleDict # containing `nn.Embedding` for `token_factors`, `token_bias`, `label_factors` and `label_bias` from pretrained L2R model
                        ):
    
    return l2r['token_factors'](sentc_dec.long()) @ l2r['label_factors'].weight.T + l2r['token_bias'](sentc_dec.long()) + l2r['label_bias'].weight.T

# %% ../nbs/01_layers.ipynb 32
class _Pay_Attention:
    def __init__(self, f, based_on): store_attr('f,based_on')
    def __call__(self, sentc): return self.f(sentc, self.based_on)

# %% ../nbs/01_layers.ipynb 33
def Linear_Attention(based_on: Module): return _Pay_Attention(_linear_attention, based_on)

# %% ../nbs/01_layers.ipynb 35
def Planted_Attention(brain: Tensor): return _Pay_Attention(_planted_attention, brain)

# %% ../nbs/01_layers.ipynb 37
class PlantedLMDecoder(Module):
    def __init__(self, 
        n_out:int, # vocab_sz 
        n_hid:int, # Number of features in encoder last layer output
        output_p:float=0.1, # Input dropout probability
        plant_wgts:dict=None, # If supplied loads `plant_wgts` into decoder
        bias:bool=True # If `False` the layer will not learn additive bias
    ):
        self.decoder = nn.Linear(n_hid, n_out, bias=bias)
        self.output_dp = RNNDropout(output_p)
        if plant_wgts: self.load_state_dict(plant_wgts)

    def forward(self, input):
        dp_inp = self.output_dp(input)
        return self.decoder(dp_inp)
        # return self.decoder(dp_inp).softmax(dim=-1).argmax(dim=-1)

# %% ../nbs/01_layers.ipynb 40
def Diffntble_Planted_Attention(l2r: nn.ModuleDict): return _Pay_Attention(_diffntble_planted_attention, l2r)

# %% ../nbs/01_layers.ipynb 42
def lincomb(t, wgts=None):
    "returns the linear combination of the dim1 of a 3d tensor of `t` based on `wgts` (if `wgts` is `None` just adds the rows)"
    if wgts is None: wgts = t.new_ones(t.size(0), 1, t.size(1))
    return torch.bmm(wgts, t) # wgts@t

# %% ../nbs/01_layers.ipynb 44
@patch
@torch.no_grad()
def topkmax(self:Tensor, k=None, dim=1):
    """
    returns softmax of the 1th dim of 3d tensor x after zeroing out values in x smaller than `k`th largest.
    If k is `None` behaves like `x.softmax(dim=dim). Intuitively, `topkmax` hedges more compared to `F.softmax``
    """
    if dim!=1: raise NotImplementedError
    k = min(k if k is not None else np.inf, self.size(dim)-1)
    kth_largest = self.sort(dim=dim, descending=True).values[:,k,:][:,None,:].repeat(1, self.size(dim), 1)
    self[self < kth_largest] = 0.
    return self.softmax(dim=1)

# %% ../nbs/01_layers.ipynb 45
def split_sort(t, sp_dim, sort_dim, sp_sz=500, **kwargs):
    if t.ndim==1: return t.sort(dim=sort_dim, **kwargs).values
    return torch.cat([s.sort(dim=sort_dim, **kwargs).values for s in torch.split(t, split_size_or_sections=sp_sz, dim=sp_dim)], dim=sp_dim)

# %% ../nbs/01_layers.ipynb 47
@patch
@torch.no_grad()
def inattention(self:Tensor, k=None, sort_dim=0, sp_dim=0):
    """
    returns `self` after zeroing out values smaller than `k`th largest in dimension `dim`.
    If k is `None` behaves like returns self.
    """
    k = min(k if k is not None else np.inf, self.size(sort_dim)-1)
    k_slice= [slice(None)]*self.ndim
    # rep = [1]*self.ndim
    k_slice[sort_dim] = k
    if len(k_slice) == 1: k_slice=k_slice[0]
    # rep[sort_dim] = self.size(sort_dim)
    kth_largest = split_sort(self, sp_dim=sp_dim, sort_dim=sort_dim, descending=True)[k_slice].unsqueeze(dim=sort_dim)#.repeat(*rep)
    # [This line wasted 6 months of deb's life :(] clone = self.detach().clone()
    self[self < kth_largest] = 0.
    return self

# %% ../nbs/01_layers.ipynb 51
from fastai.losses import CrossEntropyLossFlat
from .utils import *

# %% ../nbs/01_layers.ipynb 52
class XMLAttention(Module):
    "Compute label specific attention weights for each token in a sequence"
    def __init__(self, n_lbs, emb_sz, embed_p=0.0, plant=0.5, attn_init=(0, 0, 1), attn_damps=(1, 1, 1), static_inattn=5, diff_inattn=30, lowshot=False, unfreeze_lm_decoder=False):
        store_attr('n_lbs,emb_sz,embed_p,plant')
        self.lbs = Embedding(n_lbs, emb_sz)
        # self.lbs_weight_dp = EmbeddingDropout(self.lbs_weight, embed_p)
        self.attn = Lambda(Linear_Attention(self.lbs))
        # self.plant_wt = nn.Parameter(torch.zeros(1)) #nn.Parameter(torch.empty(1).uniform_(0,1))
        # self.splant_wt = nn.Parameter(torch.zeros(1)) #nn.Parameter(torch.empty(1).uniform_(0,1))
        # self.lin_wt = nn.Parameter(torch.ones(1)) #nn.Parameter(torch.empty(1).uniform_(0,1))
        self.wgts = nn.Parameter(tensor(attn_init).float()) # (lin_wt, plant_wt, splant_wt)
        self.damps = tensor(attn_damps).float() # (lin_wt, plant_wt, splant_wt)
        # self.k = nn.Parameter(torch.empty(1).uniform_(0, n_lbs))
        # self.k = nn.Parameter(torch.empty(1).uniform_(0, 50))
        # self.k = nn.Parameter(torch.normal(mean=torch.tensor(30.0), std=0.01))
        self.compress = nn.Linear(3*emb_sz, 400)
        init_default(self.compress)
        self.static_inattn = static_inattn
        self.diff_inattn = diff_inattn
        self.lowshot = lowshot
        self.unfreeze_lm_decoder = unfreeze_lm_decoder
    
    @property
    def attn(self): return self._attn
    @attn.setter
    def attn(self, a): self._attn = a
    
    def forward(self, inp, sentc, mask):
        # sent is the ouput of SentenceEncoder i.e., (bs, max_len tokens, nh)
        test_eqs(inp.shape, sentc.shape[:-1], mask.shape)
        if self.attn.func.f is _linear_attention:
            top_tok_attn_wgts = F.softmax(self.attn(sentc), dim=1).masked_fill(mask[:,:,None], 0) # lbl specific wts for each token (bs, max_len, n_lbs)
            lbs_cf = None
            lm_loss = 0
            return lincomb(sentc, wgts=top_tok_attn_wgts.transpose(1,2)), top_tok_attn_wgts, lbs_cf, lm_loss # for each lbl do a linear combi of all the tokens based on attn_wgts (bs, num_lbs, nh)
        elif self.attn.func.f is _planted_attention:
            attn_wgts = self.attn(inp).masked_fill(mask[:,:,None], 0)
            top_tok_attn_wgts =attn_wgts.inattention(k=15, sort_dim=1)
            top_lbs_attn_wgts = attn_wgts.clone().permute(0,2,1).inattention(k=5, sort_dim=1).permute(0,2,1).contiguous() # applying `inattention` across the lbs dim
            lbs_cf = top_lbs_attn_wgts.sum(dim=1) #shape (bs, n_lbs)
            return lincomb(sentc, wgts=top_tok_attn_wgts.transpose(1,2)), top_tok_attn_wgts, lbs_cf # for each lbl do a linear combi of all the tokens based on attn_wgts (bs, num_lbs, nh)
        elif self.attn.func.f is _diffntble_planted_attention: 
            
            # import pdb; pdb.set_trace()
            top_tok_splant_attn_wgts =  self.plant_attn(inp).masked_fill(mask[:,:,None], 0).inattention(k=self.static_inattn, sort_dim=1)
            lbs_cf = None
            # lbs_cf = self.plant_attn(inp).masked_fill(mask[:,:,None], 0).inattention(k=10, sort_dim=-1).sum(dim=1).inattention(k=15, sort_dim=-1) #shape (bs, n_lbs)
            
            top_tok_lin_attn_wgts = self.lin_attn(sentc).softmax(dim=1).masked_fill(mask[:,:,None], 0) # lbl specific wts for each token (bs, max_len, n_lbs)
            # change
            
            # import pdb; pdb.set_trace()
            
            # [real] top_tok_plant_attn_wgts = self.attn(self.lm_decoder(sentc)).masked_fill(mask[:,:,None], 0).inattention(k=self.diff_inattn, sort_dim=1).softmax(dim=1) # lbl specific wts for each token (bs, max_len, n_lbs)
            # doing research
            pred_sentc = self.lm_decoder(sentc)
            dec_sentc = pred_sentc.argmax(dim=-1)
            top_tok_plant_attn_wgts = self.attn(dec_sentc).masked_fill(mask[:,:,None], 0).inattention(k=self.diff_inattn, sort_dim=1).softmax(dim=1) # lbl specific wts for each token (bs, max_len, n_lbs)
            # import pdb; pdb.set_trace()
            lm_loss = CrossEntropyLossFlat()(pred_sentc[:,:-1], inp.masked_fill(mask,0)[:, 1:]) if self.unfreeze_lm_decoder else 0
            # doing research
            
            # top_tok_plant_attn_wgts = self.attn(self.lm_decoder(sentc)).masked_fill(mask[:,:,None], 0).inattention(k=self.k.int().item(), sort_dim=1).softmax(dim=1) # lbl specific wts for each token (bs, max_len, n_lbs)
            # top_tok_plant_attn_wgts = self.attn(self.lm_decoder(sentc)).masked_fill(mask[:,:,None], 0).softmax(dim=1) # lbl specific wts for each token (bs, max_len, n_lbs)
            # top_tok_plant_attn_wgts = self.attn(inp).masked_fill(mask[:,:,None], 0).inattention(k=30, sort_dim=1).softmax(dim=1) # lbl specific wts for each token (bs, max_len, n_lbs)
            # change
            
            # top_tok_attn_wgts = (1-self.plant)*top_tok_lin_attn_wgts + self.plant*top_tok_plant_attn_wgts
            # self.lin_wt, self.plant_wt, self.splant_wt = tensor(self.lin_wt, self.plant_wt, self.splant_wt).softmax(dim=-1)
            # self.wgts.data = self.wgts.softmax(dim=-1)
            top_tok_attn_wgts = self.wgts[0]*top_tok_lin_attn_wgts + self.wgts[1]*top_tok_plant_attn_wgts + self.wgts[2]*top_tok_splant_attn_wgts
            # import pdb; pdb.set_trace()
            # top_tok_attn_wgts = self.damps[0]*self.wgts[0]*top_tok_lin_attn_wgts + self.damps[1]*self.wgts[1]*top_tok_plant_attn_wgts + self.damps[2]*self.wgts[2]*top_tok_splant_attn_wgts
            # lbs_cf = None

            lin_comb_lin = lincomb(sentc, wgts=top_tok_lin_attn_wgts.transpose(1,2)) # real
            lin_comb_plant = lincomb(sentc, wgts=top_tok_plant_attn_wgts.transpose(1,2)) # real
            lin_comb_splant = lincomb(sentc, wgts=top_tok_splant_attn_wgts.transpose(1,2)) # real
            # lin_comb_lin_plant = self.compress(torch.cat((lin_comb_lin, lin_comb_plant, lin_comb_splant), dim=-1)).relu()
            lin_comb_lin_plant = lin_comb_lin + lin_comb_plant + lin_comb_splant # real
            
            # only for lowshot
            if self.lowshot: lin_comb_lin_plant = lincomb(sentc, wgts=top_tok_attn_wgts.transpose(1,2))

            # lin_comb_lin_plant = lincomb(sentc, wgts=top_tok_attn_wgts.transpose(1,2))
            # lin_comb_lin_plant = lin_comb_plant

            return lin_comb_lin_plant, top_tok_attn_wgts, lbs_cf, lm_loss # for each lbl do a linear combi of all the tokens based on attn_wgts (bs, num_lbs, nh)
            
        # return lincomb(sentc, wgts=top_tok_attn_wgts.transpose(1,2)), top_tok_attn_wgts, lbs_cf # for each lbl do a linear combi of all the tokens based on attn_wgts (bs, num_lbs, nh)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/10_l2r.data.load.ipynb.

# %% auto 0
__all__ = ['PreLoadTrans', 'L2RDataLoader']

# %% ../../../nbs/10_l2r.data.load.ipynb 2
from PIL import Image
from fastai.torch_imports import *
from fastai.data.load import DataLoader
from ...imports import *
from ...torch_imports import *
from ...fastai_imports import *

# %% ../../../nbs/10_l2r.data.load.ipynb 7
class PreLoadTrans():
    def __init__(self, df, qnts=None, device=None): store_attr()

    @property
    def num_lbs(self): return self._num_lbs
    @num_lbs.setter
    def num_lbs(self, l): self._num_lbs = l
    
    @property
    def num_toks(self): return self._num_toks
    @num_toks.setter
    def num_toks(self, t): self._num_toks = t
    
    def quantized_score(self, qnts=None):
        "computes quantize the rank column of `self.df['rank']` based on `qnts`"

        if qnts is None:
            qnts = torch.concat([torch.Tensor([0]), torch.logspace(torch.log10(torch.tensor(1e-4)), torch.log10(torch.tensor(1)), 100)])
        qnts = to_device(qnts, self.device)
        # import pdb; pdb.set_trace()
        # from IPython import embed; embed()
        data = qnts.new_tensor(self.df.to_numpy())
        test_eq(data.shape, (len(self.df), 3)) # dim1 is a 3 tuple (token, label, rank)
        # sort by the labels
        data = data[data[:, 1].argsort()]
        # indices of the unique labels
        # import pdb; pdb.set_trace()
            # splt_idxs = np.concatenate([torch.as_tensor(np.unique(data[:, 1].cpu().numpy(), return_index=True)[1], self.device=default_self.device(), dtype=torch.int).cpu().numpy(), array([len(data)])])
            # splt_idxs = list(splt_idxs[1:] - splt_idxs[:-1])
        splt_idxs = np.unique(data[:, 1].cpu().numpy(), return_index=True)[1][1:]
        # split by the unique labels
        data = np.split(data.cpu().numpy(), splt_idxs)
        # stacking the 0th dim with label specific data
        data = np.stack(data)
        # data = torch.as_tensor(data, dtype=qnts.dtype)
        data = qnts.new_tensor(data)
        # test_eq(data.shape, (8922, 57352, 3))
        # computing the bins based on qnts
        bins = torch.quantile(data[:, :, -1], qnts, dim=1)
        # test_eq(bins.shape, (101, 8922))
        # placing relevance scores right next to ranks
        relv_scores = bins.shape[0] - torch.searchsorted(bins.T, data[:, :, -1], right=False) # shape (8922, 57352)
        data = torch.cat((data, relv_scores.unsqueeze(-1)), dim=-1)
        # data[:, :, -1] = relv_scores
        self.num_lbs, self.num_toks = data.shape[:2]
        self.scored_toks = data
        return self.scored_toks # dim 0: labels, dim 1: 3 tuple (token, label, rank, score)

    def train_val_split(self):
        test_eq(self.scored_toks.shape, (self.num_lbs, self.num_toks, 4))
        qnts = (1 - torch.concat([tensor([0.]), torch.logspace(torch.log10(tensor(1e-4)), torch.log10(tensor(1.)), 10)])).flip(dims=(0,))
        test_eq(qnts.shape, [11])
        bins = torch.quantile(self.scored_toks[:, :, -1], qnts, dim=1)
        test_eq(bins.shape, (11, self.num_lbs))
        binned_toks = torch.searchsorted(bins.T, self.scored_toks[:, :, -1])
        test_eq(binned_toks.shape, (self.num_lbs, self.num_toks))
        # import pdb; pdb.set_trace()
        bin_nums = binned_toks[0].unique() # all the labels have same binning because we are quantizing the scores
        right_bds = bins[:, 0][bin_nums]
        left_bds = torch.concat([tensor([-torch.inf]), bins[:, 0][bin_nums[1:]-1]])
        bin_bds = torch.concat((left_bds[:,None], right_bds[:, None]), dim=1)
        bin_size = torch.bincount(binned_toks[0])
        probs = binned_toks.new_ones(binned_toks.shape).div(3.*bin_size[binned_toks])
        # probs[bin_size[binned_toks] > 4] = 2.*probs[bin_size[binned_toks] > 4] 
        # probs[bin_size[binned_toks] <= 4] = 0.001*probs[bin_size[binned_toks] <= 4] 
        probs[binned_toks < bin_nums[-2]] = 2.*probs[binned_toks < bin_nums[-2]] 
        # probs[bin_size[binned_toks] <= 4] = 0.001*probs[bin_size[binned_toks] <= 4] 
        is_valid = torch.bernoulli(probs)
        self.binned_toks, self.probs, self.is_valid, self.bin_size, self.bin_bds = binned_toks, probs, is_valid, bin_size, bin_bds
        return self.scored_toks, self.binned_toks, self.probs, self.is_valid, self.bin_size, self.bin_bds

# %% ../../../nbs/10_l2r.data.load.ipynb 8
@patch
def pad_split(self:PreLoadTrans):
    "pads the validation set for each label to the nearest multiple of 16"
    val_sizes = self.is_valid.sum(dim = -1)
    sl = 16 * torch.floor(val_sizes.max()/16) + 16
    self.val_sl = sl.int().item()
    deficit = (sl - val_sizes).int()
    top3bins = self.binned_toks[0].unique().sort(descending=True)[0][:3]
    for i,d in enumerate(deficit):
        val_idxs = torch.where(self.is_valid[i]==True)[0]
        trn_idxs = torch.where(self.is_valid[i]==False)[0]
        top3idxs = torch.where(torch.isin(self.binned_toks[i], top3bins))[0].sort()[0]
        trn_idxs = torch.as_tensor(np.setdiff1d(trn_idxs.cpu().numpy(), top3idxs.cpu().numpy()))
        more_val_idxs = trn_idxs[torch.randperm(len(trn_idxs))[:d].long()]
        self.is_valid[i][more_val_idxs] = 1
    return self.val_sl

# %% ../../../nbs/10_l2r.data.load.ipynb 9
@patch
def count_topbins(self:PreLoadTrans):
    "counts the number of top 2 bins in the val set for each label"
    top_lens = self.binned_toks.new_zeros(self.binned_toks.shape[0]) # for every label contains the number of top 2 bins (top bins are the ones which have most relevant tokens for that label) 
    top_bins = self.binned_toks[0].unique()[-2:]
    for i, binned in enumerate(self.binned_toks):
        topbin_idxs = torch.nonzero(torch.isin(binned, top_bins)).view(-1)
        val_idxs = torch.where(self.is_valid[i] == 1)[0]
        top = val_idxs[torch.where(torch.isin(val_idxs, topbin_idxs))[0]]
        top_lens[i] = top.shape[0]
    return top_lens

# %% ../../../nbs/10_l2r.data.load.ipynb 10
@patch
def datasets(self:PreLoadTrans):
    "prepare the train/val dataset from `self.scored_toks`"
    self.val_dset = self.scored_toks[self.is_valid.bool()].view(self.scored_toks.shape[0], self.val_sl, -1)
    # trn = self.scored_toks[~self.is_valid.bool()].view(self.scored_toks.shape[0], self.scored_toks.shape[1]-self.val_sl, -1) # Use this if you want to remove validation tokens
    self.trn_dset = self.scored_toks
    return self.trn_dset, self.val_dset

# %% ../../../nbs/10_l2r.data.load.ipynb 12
class L2RDataLoader(DataLoader):
    def __init__(self, *args, **kwargs):
        self.sl, self.lbs_chunks = kwargs.pop('sl', None), kwargs.pop('lbs_chunks', None)
        if self.sl is None: self.sl = 64
        if self.lbs_chunks is None: self.lbs_chunks = 4
        super().__init__(*args, **kwargs)
        
    def randomize(self):
        seed = np.random.default_rng().integers(0, 2**32-1, 1).item()
        self.rng = np.random.default_rng(seed)
    
    def shuffle_fn(self, idxs): return self.rng.permutation(idxs)

    def get_idxs(self):
        if self.n is not None: idxs = range(self.n)
        if self.shuffle: idxs = (idx for idx in self.shuffle_fn(idxs))
        return idxs
    
    def create_batch(self, start_idx):
        return self.dset[start_idx: min(start_idx+self.bs, self.dset.shape[0])]
        
    def __len__(self):
        return int(np.ceil((np.ceil(self.dataset.shape[1]/self.sl) * self.lbs_chunks)/self.bs))
    
    def before_iter(self):
        # shuffling
        # randperm = torch.randint(low=0, high=self.dataset.shape[1], size=(self.dataset.shape[1],))
        randperm_toks_idxs = torch.randperm(self.dataset.shape[1])
        self.dataset = self.dataset[:, randperm_toks_idxs]
        randperm_lbs_idxs = torch.randperm(self.dataset.shape[0])
        self.dataset = self.dataset[randperm_lbs_idxs, :]
        size_of_dim0 = torch.ceil(self.dataset.new_empty(1).fill_(self.dataset.shape[0]/self.lbs_chunks)).item()
        pad_len_dim0 = int(self.lbs_chunks * np.floor(self.dataset.shape[0]/self.lbs_chunks) + self.lbs_chunks - self.dataset.shape[0])
        self.dataset_pad = F.pad(self.dataset, (0,0,0,0,0,pad_len_dim0), value=-1)

        trn_sqs = list(torch.split(self.dataset_pad, split_size_or_sections=self.sl, dim=1))
        test_eq(len(trn_sqs), np.ceil(self.dataset_pad.shape[1]/self.sl))
        test_eq(trn_sqs[-1].shape, (self.dataset_pad.shape[0], self.dataset_pad.shape[1]%self.sl,4))
        deficit = self.sl - trn_sqs[-1].shape[1]
        if deficit: 
            test_eq(trn_sqs[-1].shape, (self.dataset_pad.shape[0], self.dataset_pad.shape[1]%self.sl,4));
            # trn_sqs[-1] = trn_sqs[-1].repeat_interleave(self.sl//trn_sqs[-1].shape[1], dim=1)
            trn_sqs[-1] = torch.cat((trn_sqs[-1],trn_sqs[-1][:, -1].unsqueeze(1).repeat_interleave(deficit, dim=1)), dim=1)
        test_eq(trn_sqs[-1].shape, (self.dataset_pad.shape[0], self.sl,4));
        
        trn_sqs = map(partial(torch.chunk, chunks=self.lbs_chunks), trn_sqs)
        trn_sqs = itertools.chain.from_iterable(trn_sqs)
        self.dset = trn_sqs # shape (size_of_one_lbs_chunk, sl, 4) 
    
    def create_batches(self, samps):
        yield from (torch.stack(btch) for btch in self.chunkify(self.dset)) # batching bs many elems from self.dset
        



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/03_text.learner.ipynb.

# %% auto 0
__all__ = ['match_collab', 'brainsplant', 'brainsplant_diffntble', 'load_collab_keys', 'TextLearner',
           'xmltext_classifier_learner']

# %% ../../nbs/03_text.learner.ipynb 1
from fastai.basics import *
from fastai.text.learner import *
from fastai.callback.rnn import *
from fastai.text.models.awdlstm import *
from fastai.text.models.core import get_text_classifier
from fastprogress.fastprogress import master_bar, progress_bar
from .models.core import *

# %% ../../nbs/03_text.learner.ipynb 7
def _get_text_vocab(dls:DataLoaders) -> list:
    "Get text vocabulary from `DataLoaders`"
    vocab = dls.vocab
    if isinstance(vocab, L): vocab = vocab[0]
    return vocab

# %% ../../nbs/03_text.learner.ipynb 8
def _get_label_vocab(dls:DataLoaders) -> list:
    "Get label vocabulary from `DataLoaders`"
    vocab = dls.vocab
    if isinstance(vocab, L): vocab = vocab[1]
    return vocab

# %% ../../nbs/03_text.learner.ipynb 9
def match_collab(
    old_wgts:dict, # Embedding weights of the colab model
    collab_vocab:dict, # Vocabulary of `token` and `label` used for colab pre-training
    lbs_vocab:list # Current labels vocabulary
) -> dict:
    "Convert the label embedding in `old_wgts` to go from `old_vocab` in colab to `lbs_vocab`"
    bias, wgts = old_wgts.get('i_bias.weight', None), old_wgts.get('i_weight.weight')
    wgts_m = wgts.mean(0)
    new_wgts = wgts.new_zeros((len(lbs_vocab), wgts.size(1)))
    if bias is not None:
        bias_m = bias.mean(0)
        new_bias = bias.new_zeros((len(lbs_vocab), 1))
    collab_lbs_vocab = collab_vocab['label']
    collab_o2i = collab_lbs_vocab.o2i if hasattr(collab_lbs_vocab, 'o2i') else {w:i for i,w in enumerate(collab_lbs_vocab)}
    missing = 0
    for i,w in enumerate(lbs_vocab):
        idx = collab_o2i.get(w, -1)
        new_wgts[i] = wgts[idx] if idx>=0 else wgts_m
        if bias is not None: new_bias[i] = bias[idx] if idx>=0 else bias_m
        if idx == -1: missing = missing + 1
    old_wgts['i_weight.weight'] = new_wgts
    if bias is not None: old_wgts['i_bias.weight'] = new_bias
    return old_wgts, missing

# %% ../../nbs/03_text.learner.ipynb 24
def _xml2brain(xml_vocab, brain_vocab, parent_bar=None):
    "Creates a mapping between the indices of the xml vocab and the brain vocab"
    pbar = progress_bar(xml_vocab, parent=parent_bar, leave=True)
    xml2brain = {i: brain_vocab.index(o) if o in brain_vocab else np.inf  for i,o in enumerate(pbar)}
    xml2brain_notfnd = [o for o in xml2brain if xml2brain[o] is np.inf]
    return xml2brain, xml2brain_notfnd

# %% ../../nbs/03_text.learner.ipynb 27
def brainsplant(xml_vocab, brain_vocab, brain, brain_bias, device=None):
    toks_lbs = 'toks lbs'.split()
    mb = master_bar(range(2))
    for i in mb:
        globals().update(dict(zip((toks_lbs[i]+'_xml2brain', toks_lbs[i]+'_notfnd'), (_xml2brain(xml_vocab[i], brain_vocab[i], parent_bar=mb)))))
        mb.write = f"Finished Loop {i}"
    xml_brain = torch.zeros(*xml_vocab.map(len)).to(default_device() if device is None else device) # initialize empty brain
    xml_lbsbias = torch.zeros(len(xml_vocab[1])).to(default_device() if device is None else device)
    toks_map = L((xml_idx, brn_idx) for xml_idx, brn_idx in toks_xml2brain.items() if brn_idx is not np.inf) 
    lbs_map = L((xml_idx, brn_idx) for xml_idx, brn_idx in lbs_xml2brain.items() if brn_idx is not np.inf) 
    xml_brain[toks_map.itemgot(0)] = brain[toks_map.itemgot(1)][:, lbs_map.itemgot(1)] # permute toks dim to match xml and brain
    xml_brain[:, lbs_map.itemgot(0)] = xml_brain.clone() # permute lbs dim to match xml and brain
    xml_lbsbias[lbs_map.itemgot(0)] = brain_bias[lbs_map.itemgot(1)].clone() # permute toks dim to match xml and brain
    return xml_brain, xml_lbsbias, toks_map, lbs_map, toks_xml2brain, lbs_xml2brain

# %% ../../nbs/03_text.learner.ipynb 41
def brainsplant_diffntble(xml_vocab, brain_vocab, l2r_wgts, device=None):
    toks_lbs = 'toks lbs'.split()
    mb = master_bar(range(2))
    for i in mb:
        globals().update(dict(zip((toks_lbs[i]+'_xml2brain', toks_lbs[i]+'_notfnd'), (_xml2brain(xml_vocab[i], brain_vocab[i], parent_bar=mb)))))
        mb.write = f"Finished Loop {i}" 
    toks_map = L((xml_idx, brn_idx) for xml_idx, brn_idx in toks_xml2brain.items() if brn_idx is not np.inf) 
    lbs_map = L((xml_idx, brn_idx) for xml_idx, brn_idx in lbs_xml2brain.items() if brn_idx is not np.inf) 
    tf_xml = torch.zeros(len(xml_vocab[0]), l2r_wgts['token_factors.weight'].shape[1]).to(default_device() if device is None else device) 
    tb_xml = torch.zeros(len(xml_vocab[0]), 1).to(default_device() if device is None else device) 
    lf_xml = torch.zeros(len(xml_vocab[1]), l2r_wgts['label_factors.weight'].shape[1]).to(default_device() if device is None else device) 
    lb_xml = torch.zeros(len(xml_vocab[1]), 1).to(default_device() if device is None else device) 
    tf_l2r, tb_l2r, lf_l2r, lb_l2r = list(l2r_wgts.values())
    tf_xml[toks_map.itemgot(0)] = tf_l2r[toks_map.itemgot(1)].clone()
    tb_xml[toks_map.itemgot(0)] = tb_l2r[toks_map.itemgot(1)].clone()
    lf_xml[lbs_map.itemgot(0)] = lf_l2r[lbs_map.itemgot(1)].clone()
    lb_xml[lbs_map.itemgot(0)] = lb_l2r[lbs_map.itemgot(1)].clone()
    # import pdb; pdb.set_trace()
    xml_wgts = {k: xml_val for k, xml_val in zip(l2r_wgts.keys(), (tf_xml, tb_xml, lf_xml, lb_xml))}
    mod_dict = nn.ModuleDict({k.split('.')[0]: nn.Embedding(*v.size()) for k,v in xml_wgts.items()}).to(default_device() if device is None else device) 
    mod_dict.load_state_dict(xml_wgts)
    return mod_dict, toks_map, lbs_map

# %% ../../nbs/03_text.learner.ipynb 58
def load_collab_keys(
    model, # Model architecture
    wgts:dict # Model weights
) -> tuple:
    "Load only collab `wgts` (`i_weight` and `i_bias`) in `model`, keeping the rest as is"
    sd = model.state_dict()
    lbs_weight, i_weight = sd.get('1.attn.lbs_weight.weight', None), wgts.get('i_weight.weight', None)
    lbs_bias, i_bias = sd.get('1.attn.lbs_weight.bias', None), wgts.get('i_bias.weight', None) 
    if lbs_weight is not None and i_weight is not None: lbs_weight.data = i_weight.data
    if lbs_bias is not None and i_bias is not None: lbs_bias.data = i_bias.data
    if '1.attn.lbs_weight_dp.emb.weight' in sd:
        sd['1.attn.lbs_weight_dp.emb.weight'] = i_weight.data.clone()
    return model.load_state_dict(sd)

# %% ../../nbs/03_text.learner.ipynb 62
from ..layers import *
from ..layers import _planted_attention, _diffntble_planted_attention

# %% ../../nbs/03_text.learner.ipynb 63
@delegates(Learner.__init__)
class TextLearner(Learner):
    "Basic class for a `Learner` in NLP."
    def __init__(self, 
        dls:DataLoaders, # Text `DataLoaders`
        model, # A standard PyTorch model
        alpha:float=2., # Param for `RNNRegularizer`
        beta:float=1., # Param for `RNNRegularizer`
        moms:tuple=(0.8,0.7,0.8), # Momentum for `Cosine Annealing Scheduler`
        **kwargs
    ):
        super().__init__(dls, model, moms=moms, **kwargs)
        self.add_cbs(rnn_cbs())

    def save_encoder(self, 
        file:str # Filename for `Encoder` 
    ):
        "Save the encoder to `file` in the model directory"
        if rank_distrib(): return # don't save if child proc
        encoder = get_model(self.model)[0]
        if hasattr(encoder, 'module'): encoder = encoder.module
        torch.save(encoder.state_dict(), join_path_file(file, self.path/self.model_dir, ext='.pth'))
    
    @delegates(save_model)
    def save(self,
        file:str, # Filename for the state_directory of the model
        **kwargs
    ):
        """
        Save model and optimizer state (if `with_opt`) to `self.path/self.model_dir/file`
        Save `self.dls.vocab` to `self.path/self.model_dir/clas_vocab.pkl`
        """
        model_file = join_path_file(file, self.path/self.model_dir, ext='.pth')
        vocab_file = join_path_file(file+'_vocab', self.path/self.model_dir, ext='.pkl')
        save_model(model_file, self.model, getattr(self, 'opt', None), **kwargs)
        save_pickle(vocab_file, self.dls.vocab)
        return model_file

    def load_encoder(self, 
        file:str, # Filename of the saved encoder 
        device:(int,str,torch.device)=None # Device used to load, defaults to `dls` device
    ):
        "Load the encoder `file` from the model directory, optionally ensuring it's on `device`"
        encoder = get_model(self.model)[0]
        if device is None: device = self.dls.device
        if hasattr(encoder, 'module'): encoder = encoder.module
        distrib_barrier()
        wgts = torch.load(join_path_file(file,self.path/self.model_dir, ext='.pth'), map_location=device)
        encoder.load_state_dict(clean_raw_keys(wgts))
        self.freeze()
        return self

    def load_pretrained(self, 
        wgts_fname:str, # Filename of saved weights 
        vocab_fname:str, # Saved vocabulary filename in pickle format
        model=None # Model to load parameters from, defaults to `Learner.model`
    ):
        "Load a pretrained model and adapt it to the data vocabulary."
        old_vocab = load_pickle(vocab_fname)
        new_vocab = _get_text_vocab(self.dls)
        distrib_barrier()
        wgts = torch.load(wgts_fname, map_location = lambda storage,loc: storage)
        if 'model' in wgts: wgts = wgts['model'] #Just in case the pretrained model was saved with an optimizer
        wgts = match_embeds(wgts, old_vocab, new_vocab)
        load_ignore_keys(self.model if model is None else model, clean_raw_keys(wgts))
        self.freeze()
        return self

    #For previous versions compatibility. Remove at release
    @delegates(load_model_text)
    def load(self, 
        file:str, # Filename of saved model 
        with_opt:bool=None, # Enable to load `Optimizer` state
        device:(int,str,torch.device)=None, # Device used to load, defaults to `dls` device
        **kwargs
    ):
        if device is None: device = self.dls.device
        if self.opt is None: self.create_opt()
        file = join_path_file(file, self.path/self.model_dir, ext='.pth')
        load_model_text(file, self.model, self.opt, device=device, strict=False, **kwargs)
        return self
    
    def load_collab(self,
        wgts_fname:str, # Filename of the saved collab model
        collab_vocab_fname:str, # Saved Vocabulary of collab labels in pickle format 
        model=None # Model to load parameters from, defaults to `Learner.model`
    ):
        "Load the label embeddings learned by collab model`, and adapt it to the label vocabulary."
        collab_vocab = load_pickle(collab_vocab_fname)
        lbs_vocab = _get_label_vocab(self.dls)
        distrib_barrier()
        wgts = torch.load(wgts_fname, map_location=lambda storage,loc: storage)
        if 'model' in wgts: wgts = wgts['model'] #Just in case the pretrained model was saved with an optimizer
        wgts, _ = match_collab(wgts, collab_vocab, lbs_vocab)
        load_collab_keys(self.model if model is None else model, wgts)
        self.freeze()
        return self

# %% ../../nbs/03_text.learner.ipynb 65
@patch
def save_decoder(self:LMLearner,
                 file:str # Filename for `Decoder`
    ):
    "Save the decoder to `file` in the model directory"
    if rank_distrib(): return # don't save if child proc
    decoder = get_model(self.model)[1]
    if hasattr(decoder, 'module'): decoder = decoder.module
    torch.save(decoder.state_dict(), join_path_file(file, self.path/self.model_dir, ext='.pth'))

# %% ../../nbs/03_text.learner.ipynb 66
@patch
def load_brain(self:TextLearner,
               file_wgts: str, # Filename of the saved attention wgts
               file_bias: str, # Filename of the saved label bias
               device:(int,str,torch.device)=None # Device used to load, defaults to `default_device()`
              ):
    """Load the pre-learnt label specific attention weights for each token from `file` located in the model directory, 
    optionally ensuring it's one `device`
    """
    brain_path = join_path_file(file_wgts, self.path/self.model_dir, ext='.pkl')
    bias_path = join_path_file(file_bias, self.path/self.model_dir, ext='.pkl')
    brain_bootstrap = torch.load(brain_path, map_location=default_device() if device is None else device)
    *brain_vocab, brain = mapt(brain_bootstrap.get, ['toks', 'lbs', 'mutual_info_jaccard'])
    brain_vocab = L(brain_vocab).map(listify)
    vocab = L(_get_text_vocab(self.dls), _get_label_vocab(self.dls)).map(listify)
    brain_bias = torch.load(bias_path, map_location=default_device() if device is None else device)
    brain_bias = brain_bias[:, :, 0].squeeze(-1)
    print("Performing brainsplant...")
    self.brain, self.lbsbias, *_ = brainsplant(vocab, brain_vocab, brain, brain_bias)
    print("Successfull!")
    # import pdb; pdb.set_trace()
    plant_attn_layer = Lambda(Planted_Attention(self.brain))
    setattr(self.model[1].pay_attn, 'attn', plant_attn_layer)
    assert self.model[1].pay_attn.attn.func.f is _planted_attention
    return self

# %% ../../nbs/03_text.learner.ipynb 67
@patch
def load_diffntble_brain(self:TextLearner,
                         file_brain:str, # Filename of the bootstrap info for l2r
                         file_l2r_wgts:str, # Filename of the pretrained l2r wgts
                         file_lm_decoder_wgts:str, # Filename of the pretrained LM decoder wgts
                         device:(int,str,torch.device)=None # # Device used to load, defaults to `default_device()`
                        ):
    """Loads the pre-learnt L2R and LM decoder wgts from `file_l2r_wgts` and ` file_lm_decoder_wgts` located in the
    model directory, optionally ensuring it's on `device`
    """
    brain_bootstrap = torch.load(join_path_file(file_brain, self.path/self.model_dir, ext='.pkl'), map_location=default_device() if device is None else device)
    brain_vocab = mapt(brain_bootstrap.get, ['toks', 'lbs'])
    brain_vocab = L(brain_vocab).map(listify)
    vocab = L(_get_text_vocab(self.dls), _get_label_vocab(self.dls)).map(listify)
    l2r_wgts = torch.load(join_path_file(file_l2r_wgts, self.path/self.model_dir, ext='.pth'), map_location=default_device() if device is None else device)
    if 'model' in l2r_wgts: l2r_wgts = l2r_wgts['model']
    print("Performing 'differentiable' brainsplant...")
    l2r, toks_map, lbs_map = brainsplant_diffntble(vocab, brain_vocab, l2r_wgts)
    print("Successfull!")
    lm_decoder_pretrained_wgts = torch.load(join_path_file(file_lm_decoder_wgts, self.path/self.model_dir, ext='.pth'), map_location=default_device() if device is None else device)
    config = awd_lstm_lm_config.copy()
    emb_sz, output_p, out_bias = map(config.get, ['emb_sz', 'output_p', 'out_bias'])
    lm_decoder = PlantedLMDecoder(len(vocab[0]), emb_sz, output_p=output_p*0.3, plant_wgts=lm_decoder_pretrained_wgts, bias=out_bias).to(default_device() if device is None else device)
    test_eq(lm_decoder.decoder.weight, lm_decoder_pretrained_wgts['decoder.weight'])
    test_eq(lm_decoder.decoder.bias, lm_decoder_pretrained_wgts['decoder.bias'])
    plant_attn_layer = Lambda(Diffntble_Planted_Attention(l2r))
    setattr(self.model[1].pay_attn, 'attn', plant_attn_layer)
    assert self.model[1].pay_attn.attn.func.f is _diffntble_planted_attention
    setattr(self.model[1].pay_attn, 'lm_decoder', lm_decoder)
    return self

# %% ../../nbs/03_text.learner.ipynb 68
@patch
def load_both(self:TextLearner,
                         file_brain:str, # Filename of the bootstrap info for l2r
                         file_bias: str, # Filename of the saved label bias
                         file_l2r_wgts:str, # Filename of the pretrained l2r wgts
                         file_lm_decoder_wgts:str, # Filename of the pretrained LM decoder wgts
                         device:(int,str,torch.device)=None # # Device used to load, defaults to `default_device()`
                        ):
    bias_path = join_path_file(file_bias, self.path/self.model_dir, ext='.pkl')
    brain_path = join_path_file(file_brain, self.path/self.model_dir, ext='.pkl')
    brain_bootstrap = torch.load(brain_path, map_location=default_device() if device is None else device)
    *brain_vocab, brain = mapt(brain_bootstrap.get, ['toks', 'lbs', 'mutual_info_jaccard'])
    brain_vocab = L(brain_vocab).map(listify)
    vocab = L(_get_text_vocab(self.dls), _get_label_vocab(self.dls)).map(listify)
    brain_bias = torch.load(bias_path, map_location=default_device() if device is None else device)
    brain_bias = brain_bias[:, :, 0].squeeze(-1)
    print("Performing static brainsplant...")
    self.brain, self.lbsbias, *_ = brainsplant(vocab, brain_vocab, brain, brain_bias)
    print("Successfull!")
    plant_attn_layer = Lambda(Planted_Attention(self.brain))
    setattr(self.model[1].pay_attn, 'plant_attn', plant_attn_layer)
    assert self.model[1].pay_attn.plant_attn.func.f is _planted_attention
    
    l2r_wgts = torch.load(join_path_file(file_l2r_wgts, self.path/self.model_dir, ext='.pth'), map_location=default_device() if device is None else device)
    if 'model' in l2r_wgts: l2r_wgts = l2r_wgts['model']
    print("Performing 'differentiable' brainsplant...")
    l2r, toks_map, lbs_map = brainsplant_diffntble(vocab, brain_vocab, l2r_wgts)
    print("Successfull!")
    lm_decoder_pretrained_wgts = torch.load(join_path_file(file_lm_decoder_wgts, self.path/self.model_dir, ext='.pth'), map_location=default_device() if device is None else device)
    config = awd_lstm_lm_config.copy()
    emb_sz, output_p, out_bias = map(config.get, ['emb_sz', 'output_p', 'out_bias'])
    lm_decoder = PlantedLMDecoder(len(vocab[0]), emb_sz, output_p=output_p*0.3, plant_wgts=lm_decoder_pretrained_wgts, bias=out_bias).to(default_device() if device is None else device)
    test_eq(lm_decoder.decoder.weight, lm_decoder_pretrained_wgts['decoder.weight'])
    test_eq(lm_decoder.decoder.bias, lm_decoder_pretrained_wgts['decoder.bias'])
    plant_attn_layer = Lambda(Diffntble_Planted_Attention(l2r))
    
    lin_attn = getattr(self.model[1].pay_attn, 'attn')
    setattr(self.model[1].pay_attn, 'attn', plant_attn_layer)
    setattr(self.model[1].pay_attn, 'l2r', l2r)
    setattr(self.model[1].pay_attn, 'lin_attn', lin_attn)
    assert self.model[1].pay_attn.attn.func.f is _diffntble_planted_attention
    setattr(self.model[1].pay_attn, 'lm_decoder', lm_decoder)
    
    return self 

# %% ../../nbs/03_text.learner.ipynb 71
from .models.core import _model_meta 

# %% ../../nbs/03_text.learner.ipynb 72
@delegates(Learner.__init__)
def xmltext_classifier_learner(dls, arch, seq_len=72, config=None, backwards=False, pretrained=True, collab=False, drop_mult=0.5, n_out=None,
                           lin_ftrs=None, ps=None, max_len=72*20, y_range=None, splitter=None, running_decoder=True, plant=0.5, attn_init=(0, 0, 1), static_inattn=5, diff_inattn=30, lowshot=False, unfreeze_lm_decoder=False, **kwargs):
    "Create a `Learner` with a text classifier from `dls` and `arch`."
    vocab = _get_text_vocab(dls)
    if n_out is None: n_out = get_c(dls)
    assert n_out, "`n_out` is not defined, and could not be inferred from the data, set `dls.c` or pass `n_out`"
    model = get_xmltext_classifier2(arch, len(vocab), n_out, seq_len=seq_len, config=config, y_range=y_range,
                                drop_mult=drop_mult, max_len=max_len, running_decoder=running_decoder, plant=plant, attn_init=attn_init,
                                static_inattn=static_inattn, diff_inattn=diff_inattn, lowshot=lowshot, unfreeze_lm_decoder=unfreeze_lm_decoder)
    # model = get_xmltext_classifier(arch, len(vocab), n_out, seq_len=seq_len, config=config, y_range=y_range,
                                # drop_mult=drop_mult, max_len=max_len)
    meta = _model_meta[arch]
    learn = TextLearner(dls, model, splitter=splitter if splitter is not None else meta['split_clas'], **kwargs)
    url = 'url_bwd' if backwards else 'url'
    if pretrained:
        if url not in meta:
            warn("There are no pretrained weights for that architecture yet!")
            return learn
        model_path = untar_data(meta[url], c_key='model')
        try: fnames = [list(model_path.glob(f'*.{ext}'))[0] for ext in ['pth', 'pkl']]
        except IndexError: print(f'The model in {model_path} is incomplete, download again'); raise
        learn = learn.load_pretrained(*fnames, model=learn.model[0])
    if collab:
        try: fnames = [list(learn.path.glob(f'**/collab/*collab*.{ext}'))[0] for ext in ['pth', 'pkl']]
        except IndexError: print(f'The collab model in {learn.path} is incomplete, re-train it!'); raise
        learn = learn.load_colab(*fnames, model=learn.model[1])
    learn.freeze()
    return learn   
